/*
 * Copyright (c) 2023 XMLmind Software. All rights reserved.
 *
 * Author: Hussein Shafie
 *
 * This file is part of several XMLmind projects.
 * For conditions of distribution and use, see the accompanying LEGAL.txt file.
 */
package com.xmlmind.util;

import java.util.concurrent.atomic.AtomicBoolean;

/**
 * A {@link Thread} in which deprecated, marked for removal, 
 * method {@link Thread#stop} has been "replaced" by 
 * {@link StoppableThread#requestStop}, {@link StoppableThread#stopRequested} 
 * and {@link StoppableThread#kill}.
 */
public class StoppableThread extends Thread {
    private final AtomicBoolean stopFlag = new AtomicBoolean(false);

    // -----------------------------------------------------------------------
    
    public StoppableThread() {
        super();
    }
    
    public StoppableThread(Runnable task) {
        super(task);
    }
    
    public StoppableThread(String name) {
        super(name);
    }
    
    public StoppableThread(Runnable task, String name) {
        super(task, name);
    }

    public StoppableThread(ThreadGroup group, Runnable task) {
        super(group, task);
    }
    
    public StoppableThread(ThreadGroup group, String name) {
        super(group, name);
    }
    
    public StoppableThread(ThreadGroup group, Runnable task, String name) {
        super(group, task, name);
    }

    public StoppableThread(ThreadGroup group, Runnable task, String name,
                           long stackSize) {
        super(group, task, name, stackSize);
    }

    @Override
    public void start() {
        stopFlag.set(false);
        super.start();
    }

    /**
     * Request this thread to stop and interrupt it.
     */
    public void requestStop() {
        stopFlag.set(true);
        this.interrupt();
    }

    /**
     * Tests wether {@link #requestStop} has already been invoked.
     */
    public boolean stopRequested() {
        return stopFlag.get();
    }

    /**
     * Tests wether {@link #requestStop} has already been invoked for
     * {@link Thread#currentThread}.
     */
    public static boolean stopIsRequested() {
        Thread current = Thread.currentThread();
        if (current instanceof StoppableThread) {
            return ((StoppableThread) current).stopRequested();
        } else {
            return false;
        }
    }

    /**
     * Replaces deprecated, marked for removal, method {@link Thread#stop}.
     * Unlikely to have any effect on this thread.
     */
    @SuppressWarnings({"deprecation", "removal"})
    public void kill() {
        try {
            this.stop();
        } catch (Throwable ignored) {}
    }
}
