/*
 * Copyright (c) 2017-2023 XMLmind Software. All rights reserved.
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind Web Help Compiler project.
 * For conditions of distribution and use, see the accompanying LEGAL.txt file.
 */
package com.xmlmind.whc;

import java.io.IOException;
import java.io.File;
import java.io.InputStream;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.UUID;
import java.util.Locale;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.Stack;
import java.util.Arrays;
import java.util.Collections;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.text.Collator;
import static javax.xml.XMLConstants.XML_NS_URI;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import com.xmlmind.util.ThrowableUtil;
import com.xmlmind.util.LocaleUtil;
import com.xmlmind.util.StringList;
import com.xmlmind.util.StringUtil;
import com.xmlmind.util.FileUtil;
import com.xmlmind.util.URIComponent;
import com.xmlmind.util.URLUtil;
import com.xmlmind.util.LoadText;
import com.xmlmind.util.XMLText;
import com.xmlmind.util.Console;
import static com.xmlmind.whc.NamespaceHelper.XHTML_NS_URI;

/**
 * The Web Help Compiler.
 *
 * <p>This class is not thread-safe. 
 * That is, create and use one instance per thread.
 * However an instance may be reused by invoking 
 * {@link #compile(URL[], URL, URL, File)} several times.
 */
public final class Compiler {
    /**
     * The version number of this Compiler.
     */
    public static final String VERSION = "3.5.1";

    // -----------------------------------------------------------------------

    private static final URL[] templateManifestURL = new URL[1];

    /**
     * Specifies the location of <tt>whc_template/manifest.txt</tt>, 
     * the manifest file listing all the files used by the compiler 
     * to ``decorate'' an XHTML page.
     * <p>When this location has not been explicitly specified, it is searched 
     * in the following location:
     * <ol>
     * <li>resource <tt>/whc_template/manifest.txt</tt> 
     * if found in <tt>whc.jar</tt>,
     * <li><tt><i>whc_jar_dir</i>/whc_template/manifest.txt</tt>,
     * <li><tt><i>whc_jar_dir</i>/../whc_template/manifest.txt</tt>.
     * </ol>
     *
     * @param url location of <tt>whc_template/manifest.txt</tt>.
     * May be <code>null</code> in which case 
     * <tt>whc_template/manifest.txt</tt> will be searched
     * in the above locations.
     * @see #getTemplateManifestURL
     */
    public static void setTemplateManifestURL(URL url) {
        synchronized (templateManifestURL) {
            templateManifestURL[0] = url;
        }
    }

    /**
     * Returns the location of <tt>whc_template/manifest.txt</tt>,
     * May return <code>null</code>.
     *
     * @see #setTemplateManifestURL
     */
    public static URL getTemplateManifestURL() {
        synchronized (templateManifestURL) {
            if (templateManifestURL[0] == null) {
                templateManifestURL[0] = findTemplateManifestURL();
            }
            return templateManifestURL[0];
        }
    }

    private static URL findTemplateManifestURL() {
        URL url = Compiler.class.getResource("/whc_template/manifest.txt");
        if (url == null) {
            File jarFile = findJarFile();
            if (jarFile != null) {
                File jarDir = jarFile.getParentFile();

                File manifestFile = 
                    new File(jarDir, 
                             "whc_template" + File.separator + "manifest.txt");
                if (!manifestFile.isFile()) {
                    manifestFile = 
                        new File(jarDir, 
                                 ".." + File.separator + "whc_template" + 
                                 File.separator + "manifest.txt");
                    if (!manifestFile.isFile()) {
                        manifestFile = null;
                    }
                }

                if (manifestFile != null) {
                    url = FileUtil.fileToURL(manifestFile);
                }
            }
        }

        return url;
    }

    private static File findJarFile() {
        String classPath = System.getProperty("java.class.path");
        if (classPath == null) {
            return null;
        }

        File jarFile = null;

        final String[] paths = StringUtil.split(classPath, 
                                                File.pathSeparatorChar);
        for (String path : paths) {
            path = path.trim();
            if (path.endsWith(".jar")) {
                File file = null;
                try {
                    file = (new File(path)).getCanonicalFile();
                    if (!file.isFile()) {
                        file = null;
                    }
                } catch (IOException ignored) {}

                if (file != null && file.getName().startsWith("whc")) {
                    jarFile = file;
                    // Found.
                    break;
                }
            }
        }

        return jarFile;
    }

    // -----------------------------------------------------------------------

    /**
     *
     *
     * @see #setUILanguage
     */
    public static final String DOCUMENT_LANGUAGE = "document";

    /**
     *
     * @see #setUILanguage
     */
    public static final String BROWSER_LANGUAGE = "browser";

    private Console console;

    private boolean verbose;

    private String layout;
    private boolean localJQuery;
    private String jQuery;
    private URL userHeader;
    private URL userFooter;
    private URL userCSS;
    private File userResources;
    private String defaultLang;
    private boolean collapseTOC;
    private boolean useStemming;
    private boolean indexNumbers;
    private String idMapTitle;
    private String idMapHref;
    private String idMapIdPrefix;
    private boolean responsiveUI;
    private String uiLanguage;
    private HashMap<String,String> themeProperties;
    private boolean embedCSS;
    private boolean testMode;
    private boolean noRole;

    private Document userHeaderDoc;
    private Document userFooterDoc;

    // -----------------------------------------------------------------------

    /**
     * Constructs a compiler.
     *
     * @param console the console used to display error, warning 
     * and progress ({@link #setVerbose}) messages. 
     * May be <code>null</code> in which case these messages are printed
     * <code>System.err</code> and <code>System.out</code>.
     */
    public Compiler(Console console) {
        this.console = console;

        setLayout(null);
        setDefaultLanguage(null);
        useStemming = true;
        responsiveUI = true;
        setUILanguage(BROWSER_LANGUAGE);
        themeProperties = new HashMap<String,String>();
    }

    /**
     * Returns the console used to display error, warning and 
     * progress messages. May return <code>null</code>.
     */
    public Console getConsole() {
        return console;
    }

    /**
     * Specifies the verbosity of this compiler.
     * <p>By default the compiler is not verbose.
     *
     * @param verbose if <code>true</code>, display progress messages
     * on the console.
     * @see #isVerbose
     */
    public void setVerbose(boolean verbose) {
        this.verbose = verbose;
    }

    /**
     * Returns the verbosity of this compiler.
     *
     * @see #setVerbose
     */
    public boolean isVerbose() {
        return verbose;
    }

    /**
     * Selects a layout for the generated Web Help.
     *
     * @param layout layout name. 
     * For now, only "<tt>classic</tt>", "<tt>simple</tt>" and 
     * "<tt>corporate</tt>" layouts are supported.
     * May be <code>null</code> in which case "<tt>classic</tt>" is used.
     * @see #getLayout
     */
    public void setLayout(String layout) {
        if (layout == null || (layout = layout.trim()).length() == 0) {
            layout = "classic";
        }
        this.layout = layout;
    }

    /**
     * Returns the selected selected for the generated Web Help.
     *
     * @see #setLayout
     */
    public String getLayout() {
        return layout;
    }

    /**
     * Specifies whether the jQuery files must be copied to 
     * the local resource directory created by this compiler.
     * <p>Default: Not local.
     * jQuery files are downloaded from the Web by the Web Help.
     *
     * @param local <code>true</code> the jQuery files must be local to
     * the Web Help; <code>false</code> if they are to be  downloaded 
     * from the Web by the Web Help
     * @see #isLocalJQuery
     */
    public void setLocalJQuery(boolean local) {
        this.localJQuery = local;
    }

    /**
     * Returns <code>true</code> if the jQuery files must be copied to
     * the local resource directory created by this compiler; 
     * <code>false</code> otherwise.
     *
     * @see #setLocalJQuery
     */
    public boolean isLocalJQuery() {
        return localJQuery;
    }

    /**
     * Specifies the location of the JavaScript file containing jQuery.
     * <p>By default the compiler references a file found on the Google CDN.
     *
     * @param location relative or absolute URI of 
     * the JavaScript file containing jQuery.
     * @see #getJQuery
     */
    public void setJQuery(String location) {
        jQuery = location;
    }

    /**
     * Returns the location of the JavaScript file containing jQuery.
     * May return <code>null</code>.
     *
     * @see #setJQuery
     */
    public String getJQuery() {
        return jQuery;
    }

    /**
     * Specifies the user's header which is to be added to an 
     * XHTML page decorated by the compiler.
     *
     * @param url location of an XHTML page.
     * The content of its <tt>body</tt> element is inserted as is in
     * the <tt>div id="wh-header"</tt> found in an XHTML page 
     * decorated by the compiler.
     * @see #getUserHeader
     */
    public void setUserHeader(URL url) {
        userHeader = url;
        userHeaderDoc = null;
    }

    /**
     * Returns the user's header to be added to an 
     * XHTML page decorated by the compiler.
     * May return <code>null</code>.
     *
     * @see #setUserHeader
     */
    public URL getUserHeader() {
        return userHeader;
    }

    /**
     * Specifies the user's footer which is to be added to an 
     * XHTML page decorated by the compiler.
     *
     * @param url location of an XHTML page.
     * The content of its <tt>body</tt> element is inserted as is in
     * the <tt>div id="wh-footer"</tt> found in an XHTML page 
     * decorated by the compiler.
     * @see #getUserFooter
     */
    public void setUserFooter(URL url) {
        userFooter = url;
        userFooterDoc = null;
    }

    /**
     * Returns the user's footer to be added to an 
     * XHTML page decorated by the compiler.
     * May return <code>null</code>.
     *
     * @see #setUserFooter
     */
    public URL getUserFooter() {
        return userFooter;
    }

    /**
     * Specifies the user's CSS stylesheet which is to be added to an 
     * XHTML page decorated by the compiler.
     * <p>This file is copied to <tt><i>output_directory</i>/_wh/user/</tt>.
     *
     * @param url location of a CSS stylesheet
     * @see #getUserCSS
     * @see #setUserResources
     */
    public void setUserCSS(URL url) {
        userCSS = url;
    }

    /**
     * Returns the user's CSS stylesheet to be added to an 
     * XHTML page decorated by the compiler.
     * May return <code>null</code>.
     *
     * @see #setUserCSS
     */
    public URL getUserCSS() {
        return userCSS;
    }

    /**
     * Specifies a user's resource directory which is to be recursively 
     * copied to <tt><i>output_directory</i>/_wh/user/</tt>.
     * This directory typically contains image files referenced
     * by the user's header, footer or CSS stylesheet.
     *
     * @param dir a directory to be recursively copied
     * to <tt><i>output_directory</i>/_wh/user/</tt>.
     * @see #getUserResources
     * @see #getUserHeader
     * @see #setUserFooter
     * @see #setUserCSS
     */
    public void setUserResources(File dir) {
        userResources = dir;
    }

    /**
     * Returns the user's resource directory which is recursively copied to
     * <tt><i>output_directiry</i>/_wh/user/</tt>.
     * May return <code>null</code>.
     *
     * @see #setUserResources
     */
    public File getUserResources() {
        return userResources;
    }

    /**
     * Specifies the language used when the XHTML page being parsed does not
     * specify a <tt>xml:lang</tt> or <tt>lang</tt> attribute 
     * on its <tt>html</tt> root element.
     * <p>This language is needed by the full-text indexer component of this
     * compiler.
     *
     * @param lang the default language of an XHTML page.
     * A language code conforming 
     * <a href="http://www.ietf.org/rfc/rfc3066.txt">RFC 3066</a>. 
     * Examples: "<tt>de</tt>", "<tt>fr-CA</tt>".
     * May be <code>null</code> in which case "<tt>en</tt>" is used.
     * @see #getDefaultLanguage
     */
    public void setDefaultLanguage(String lang) {
        if (lang == null || (lang = lang.trim()).length() == 0) {
            lang = "en";
        }
        defaultLang = lang;
    }

    /**
     * Returns the language used when the XHTML page being parsed does not
     * specify a <tt>xml:lang</tt> or <tt>lang</tt> attribute 
     * on its <tt>html</tt> root element.
     *
     * @see #setDefaultLanguage
     */
    public String getDefaultLanguage() {
        return defaultLang;
    }

    /**
     * Specifies whether the TOC should be initially collapsed.
     * <p>By default, the TOC is initially expanded.
     *
     * @param collapse if <code>true</code>, the TOC should be initially 
     * collapsed.
     * @see #getCollapseTOC
     */
    public void setCollapseTOC(boolean collapse) {
        collapseTOC = collapse;
    }

    /**
     * Returns <code>true</code> if the TOC is to be initially 
     * collapsed; otherwise returns <code>false</code>.
     *
     * @see #setCollapseTOC
     */
    public boolean getCollapseTOC() {
        return collapseTOC;
    }

    /**
     * Specifies whether stemming should be used to implement 
     * the search by contents facility.
     * <p>By default, stemming must be used whenever possible, that is, when
     * the main language of the document can be determined and when this main 
     * language is one of: danish, dutch, english, finnish, french, german, 
     * hungarian, italian, norwegian, portuguese, russian, spanish, 
     * swedish, romanian, turkish.
     *
     * @param stemming if <code>true</code>, stemming is to implement 
     * the search by contents facility
     * @see #getUseStemming
     * @see #setDefaultLanguage
     */
    public void setUseStemming(boolean stemming) {
        useStemming = stemming;
    }

    /**
     * Returns <code>true</code> if stemming is to be used to implement 
     * the search by contents facility; otherwise returns <code>false</code>.
     *
     * @see #setUseStemming
     */
    public boolean getUseStemming() {
        return useStemming;
    }

    /**
     * Specifies whether words looking like numbers are to be indexed.
     * <p>By default, words looking like numbers are not indexed.
     * <p>Examples of such words: <tt>3.14</tt>, <tt>3,14</tt>, 
     * <tt>3times4equals12</tt>, <tt>+1</tt>, <tt>-1.0</tt>, <tt>3px</tt>, 
     * <tt>1,2cm</tt>, <tt>100%</tt>, <tt>1.0E+6</tt>, <tt>1,000.00$</tt>.
     *
     * @see #getIndexNumbers
     */
    public void setIndexNumbers(boolean index) {
        indexNumbers = index;
    }

    /**
     * Returns <code>true</code> if words looking like numbers 
     * are to be indexed; otherwise returns <code>false</code>.
     *
     * @see #setIndexNumbers
     */
    public boolean getIndexNumbers() {
        return indexNumbers;
    }

    /**
     * Instructs this compiler to save an ID map to 
     * <tt><i>output_dir</i>/_wh/idMap.xml</tt>.
     *
     * @param title the value of the <code>title</code> attribute
     * of the generated <code>idMap</code> root element.
     * <p>This value is the title of the document for which 
     * a Web Help has been generated.
     * <p>May be empty or <code>null</code> in which case an ID map 
     * is not generated (which is the default behavior).
     * @see #setIdMapHref
     * @see #getIdMapTitle
     */
    public void setIdMapTitle(String title) {
        if (title != null && 
            (title = XMLText.collapseWhiteSpace(title)).length() == 0) {
            title = null;
        }
        idMapTitle = title;
    }

    /**
     * Returns the <code>title</code> value specified 
     * using {@link #setIdMapTitle}. May return <code>null</code>.
     */
    public String getIdMapTitle() {
        return idMapTitle;
    }

    /**
     * Instructs this compiler to save an ID map to 
     * <tt><i>output_dir</i>/_wh/idMap.xml</tt>.
     *
     * @param href the value of the <code>href</code> attribute
     * of the generated <code>idMap</code> root element. 
     * <p>This value should point to the HTML file which is the "entry point" 
     * of the generated Web Help.
     * <p>May be empty or <code>null</code> in which case an ID map 
     * is not generated (which is the default behavior).
     * <p>Note that this parameter is the absolute or relative URL of 
     * a file or directory. 
     * Directory example: specify <tt>My%20Doc/index.html</tt>, 
     * not <tt>My Doc\index.html</tt>.
     * @see #setIdMapTitle
     * @see #getIdMapHref
     */
    public void setIdMapHref(String href) {
        if (href != null && (href = href.trim()).length() == 0) {
            href = null;
        }
        idMapHref = href;
    }

    /**
     * Returns the <code>href</code> value specified 
     * using {@link #setIdMapHref}. May return <code>null</code>.
     */
    public String getIdMapHref() {
        return idMapHref;
    }

    /**
     * In addition to the IDs of "formal elements" (sections, figures, 
     * tables, etc) which are automatically added to the ID map, 
     * also add the IDs starting with specified prefix.
     *
     * @param prefix prefix of extra IDs to be added to the ID map.
     * @see #setIdMapTitle
     * @see #setIdMapHref
     */
    public void setIdMapIdPrefix(String prefix) {
        if (prefix != null && (prefix = prefix.trim()).length() == 0) {
            prefix = null;
        }
        idMapIdPrefix = prefix;
    }

    /**
     * Returns the ID prefix specified using {@link #setIdMapIdPrefix}. 
     * May return <code>null</code>.
     */
    public String getIdMapIdPrefix() {
        return idMapIdPrefix;
    }

    /**
     * Specifies whether the UI of the generated Web Help 
     * is to be "responsive". Default: <code>true</code>.
     *
     * @see #getResponsiveUI
     */
    public void setResponsiveUI(boolean responsive) {
        responsiveUI = responsive;
    }

    /**
     * Returns <code>true</code> if the UI of the generated Web Help 
     * is to be "responsive"; <code>false</code> otherwise.
     *
     * @see #setResponsiveUI
     */
    public boolean getResponsiveUI() {
        return responsiveUI;
    }

    /**
     * Specifies the language to be used for the messages of 
     * the generated Web Help. Default: {@link #BROWSER_LANGUAGE}.
     *
     * @param lang the language to be used. An ISO 639-1 language code 
     * (e.g. "<code>de</code>", "<code>fr-CH</code>") or 
     * {@link #DOCUMENT_LANGUAGE} or {@link #BROWSER_LANGUAGE} or 
     * <code>null</code> (which is equivalent to <code>BROWSER_LANGUAGE</code>).
     * @see #getUILanguage
     */
    public void setUILanguage(String lang) {
        if (lang == null || (lang = lang.trim()).length() == 0) {
            lang = BROWSER_LANGUAGE;
        }

        uiLanguage = lang;
    }

    /**
     * Returns the language to be used for the messages of 
     * the generated Web Help. 
     *
     * @see #setUILanguage
     */
    public String getUILanguage() {
        return uiLanguage;
    }

    /**
     * Override a CSS theme property with specified values.
     *
     * @param name name of a custom CSS property 
     * used by the current Web Help "theme". 
     * Example: <code>--navigation-width</code>.
     * (<i><tt>name</tt></i> always starts with "<code>--</code>".)
     * @param value value of the <i><tt>name</tt></i> custom CSS property. 
     * Example: <code>25%</code>.
     * @see #getThemeProperties
     * @see #inheritFontAndColors
     */
    public void setThemeProperty(String name, String value) {
        if (value == null) {
            themeProperties.remove(name);
        } else {
            themeProperties.put(name, value);
        }
    }

    /**
     * Returns a map containing the CSS theme properties specified using
     * {@link #setThemeProperty}.
     */
    public Map<String,String> getThemeProperties() {
        return themeProperties;
    }

    /**
     * A shorthand invoking {@link #setThemeProperty} to set all the
     * CSS theme properties specifying the font and colors of 
     * the navigation pane to "<code>inherit</code>".
     */
    public void inheritFontAndColors(boolean inherit) {
        // Cannot add "--navigation-line-height" here because of 
        // "--navigation-item-height".
        final String[] names = {
            "--navigation-font-family",
            "--navigation-font-size",
            "--navigation-color",
            "--navigation-background-color"
        };

        for (String name : names) {
            setThemeProperty(name, inherit? "inherit" : null);
        }
    }

    /**
     * Parses specified parameters invoking {@link #setUILanguage}, 
     * {@link #setJQuery}, ..., {@link #setCollapseTOC}, as needed to.
     *
     * @param parameters a list of parameter name/parameter value pairs.
     * A parameter may be prefixed by "<tt>wh-</tt>", 
     * in which case this prefixed is stripped.
     * @return <code>true</code> if the parameters have been successfully 
     * parsed; <code>false</code> otherwise (in which case error messages 
     * are displayed on the console).
     */
    public boolean parseParameters(String[] parameters) {
        return parseParameters(parameters, null);
    }

    /**
     * For internal use only. Not part of the public, documented, API. 
     */
    public boolean parseParameters(String[] parameters, URL baseURL) {
        final int count = (parameters.length/2)*2;
        for (int i = 0; i < count; i += 2) {
            String name = parameters[i];
            if (name == null || (name = name.trim()).length() == 0) {
                continue;
            }

            String value = parameters[i+1];
            if (value == null || (value = value.trim()).length() == 0) {
                continue;
            }

            boolean invalidValue = false;

            if (name.startsWith("wh-")) {
                name = name.substring(3);
            }

            if ("layout".equals(name)) {
                setLayout(value);
            } else if ("local-jquery".equals(name)) {
                setLocalJQuery(parseBoolean(value));
            } else if ("jquery".equals(name)) {
                setJQuery(value);
            } else if ("user-header".equals(name)) {
                URL url = URLUtil.urlOrFile(value, /*checkAbsolute*/ false, 
                                            /*allowDir*/ false, baseURL);
                if (url == null) {
                    invalidValue = true;
                } else {
                    setUserHeader(url);
                }
            } else if ("user-footer".equals(name)) {
                URL url = URLUtil.urlOrFile(value, /*checkAbsolute*/ false, 
                                            /*allowDir*/ false, baseURL);
                if (url == null) {
                    invalidValue = true;
                } else {
                    setUserFooter(url);
                }
            } else if ("user-css".equals(name)) {
                URL url = URLUtil.urlOrFile(value, /*checkAbsolute*/ false, 
                                            /*allowDir*/ false, baseURL);
                if (url == null) {
                    invalidValue = true;
                } else {
                    setUserCSS(url);
                }
            } else if ("user-resources".equals(name)) {
                URL url = URLUtil.urlOrFile(value, /*checkAbsolute*/ false, 
                                            /*allowDir*/ true, baseURL);
                if (url == null) {
                    invalidValue = true;
                } else {
                    File dir = URLUtil.urlToFile(url);
                    if (dir == null || !dir.isDirectory()) {
                        invalidValue = true;
                    } else {
                        setUserResources(dir);
                    }
                }
            } else if ("default-language".equals(name)) {
                setDefaultLanguage(value);
            } else if ("collapse-toc".equals(name)) {
                setCollapseTOC(parseBoolean(value));
            } else if ("use-stemming".equals(name)) {
                setUseStemming(parseBoolean(value));
            } else if ("index-numbers".equals(name)) {
                setIndexNumbers(parseBoolean(value));
            } else if ("id-map-title".equals(name)) {
                setIdMapTitle(value);
            } else if ("id-map-href".equals(name)) {
                setIdMapHref(value);
            } else if ("id-map-id-prefix".equals(name)) {
                setIdMapIdPrefix(value);
            } else if ("responsive-ui".equals(name)) {
                setResponsiveUI(parseBoolean(value));
            } else if ("ui-language".equals(name)) {
                setUILanguage(value);
            } else if ("inherit-font-and-colors".equals(name)) {
                inheritFontAndColors(parseBoolean(value));
            } else if ("embed-css".equals(name)) {
                // Undocumented parameter. Needed to workaround a Java 15
                // javafx.scene.web.WebEngine bug JDK-8253696.
                embedCSS = parseBoolean(value);
            } else if ("test-mode".equals(name)) {
                // Undocumented parameter. Only used to test whc.
                testMode = parseBoolean(value);
            } else if ("no-role".equals(name)) {
                // Undocumented parameter. May be useful when testing
                // (e.g. validation).
                noRole = parseBoolean(value);
            } else {
                if (name.startsWith("-")) {
                    if (!name.startsWith("--")) {
                        // Accept both "-foo" and "--foo".
                        name = "-" + name;
                    }

                    setThemeProperty(name, value);
                } else {
                    error(Msg.msg("unknownParameter", name));
                    return false;
                }
            }

            if (invalidValue) {
                error(Msg.msg("invalidParameter", value, name));
                return false;
            }
        }
        return true;
    }

    private static boolean parseBoolean(String value) {
        return ("yes".equals(value) || 
                "true".equals(value)|| 
                "on".equals(value) || 
                "1".equals(value));
    }

    /**
     * Convenience method equivalent to 
     * {@link #compile(URL[], URL, URL, File)}, except that the arguments 
     * are <code>File</code>s and not <code>URL</code>s.
     */
    public boolean compile(File[] xhtmlFiles, File tocFile, File indexFile,
                           File outDir) 
        throws IOException {
        final int count = xhtmlFiles.length;
        URL[] xhtmlURLs = new URL[count];
        for (int i = 0; i < count; ++i) {
            xhtmlURLs[i] = FileUtil.fileToURL(xhtmlFiles[i]);
        }

        URL tocURL = null;
        if (tocFile != null) {
            tocURL = FileUtil.fileToURL(tocFile);
        }

        URL indexURL = null;
        if (indexFile != null) {
            indexURL = FileUtil.fileToURL(indexFile);
        }

        return compile(xhtmlURLs, tocURL, indexURL, outDir);
    }

    /**
     * Compiles a Web Help application by ``decorating'' specified XHTML pages.
     * <p>Decorating means: adding a few extra XHTML elements to each page.
     * These elements typically point to CSS, image and JavaScript resources
     * found in the <tt><i>output_directory</i>/_wh/</tt> directory.
     *
     * @param xhtmlURLs the list of XHTML input files to be ``decorated'' by
     * the compiler.
     * @param tocURL the XML file specifying the entries of 
     * the <b>Table Of Contents</b>. 
     * May be <code>null</code> in which case a simple TOC
     * comprising the list of the titles of the XHTML input files
     * is automatically used to populate the <b>Contents</b> tab.
     * <p>This XML file must conform to RELAX NG schema <tt>toc.rnc</tt>.
     * @param indexURL the XML file specifying the entries of the <b>Index</b>.
     * May be <code>null</code> in which case the <b>Index</b> tab 
     * is not created.
     * <p>This XML file must conform to RELAX NG schema <tt>index.rnc</tt>.
     * @param outDir the output directory. May be <code>null</code>
     * in which case the XHTML input files are modified in place.
     * <p>When the XHTML input files are decorated in place, they must all
     * be found in the same directory.
     * @return <code>true</code> if the Web Help application has 
     * been successfully generated; 
     * <code>false</code> otherwise (in which case error messages 
     * are displayed on the console).
     * @exception IOException if an I/O error occurs while reading or 
     * writing files.
     */
    public boolean compile(URL[] xhtmlURLs, URL tocURL, URL indexURL,
                           File outDir) 
        throws IOException {
        final int count = xhtmlURLs.length;
        if (count == 0) {
            error(Msg.msg("noXHTMLFiles"));
            return false;
        }

        // Create output directory ---

        if (outDir == null) {
            outDir = getOutDir(xhtmlURLs);
        }
        if (outDir == null) {
            return false;
        }
        if (!outDir.isDirectory()) {
            error(Msg.msg("noOutDir", outDir));
            return false;
        }

        File whDir = new File(outDir, "_wh");
        if (whDir.exists()) {
            FileUtil.deleteFileOrDir(whDir);
        }
        FileUtil.checkedMkdir(whDir);

        String extraCSS = copyExtraCSS(whDir);

        // Find page template and its resources ---

        String layout = getLayout();
        ArrayList<String> layoutList = new ArrayList<String>();
        ArrayList<URL> resourceList = new ArrayList<URL>();
        URL pageTemplateURL = 
            listTemplateResources(getTemplateManifestURL(), layout, 
                                  layoutList, resourceList);
        if (pageTemplateURL == null) {
            error(Msg.msg("cannotFindResource", 
                          "whc_template/" + layout + ".html"));
            return false;
        }
        if (resourceList.isEmpty()) {
            error(Msg.msg("cannotFindResource", "whc_template/_wh/*.{css,js}"));
            return false;
        }

        // Load and prepare page template ---

        info(Msg.msg("loadingPageTemplate", URLUtil.toLabel(pageTemplateURL)));

        // Ignore empty Index file.
        Document indexDoc = null;
        if (indexURL != null) {
            indexDoc = checkEmptyIndex(indexURL);
            if (indexDoc == null) {
                indexURL = null;
            }
        }

        boolean[] useIndex = new boolean[] { (indexDoc != null) };
        PageTemplateInfo pageTemplateInfo = 
            preparePageTemplate(pageTemplateURL, whDir, extraCSS, useIndex);
        if (pageTemplateInfo == null) {
            return false;
        }
        if (!useIndex[0]) {
            // The "simple" layout does not use the Index if any.
            indexDoc = null;
        }

        // Process XHTML input files ---

        ArrayList<Object> idMap = null;
        HashSet<String> collectedIds = null;
        if (idMapTitle != null && idMapHref != null) {
            idMap = new ArrayList<Object>();
            collectedIds = new HashSet<String>();
        }

        String docLang = null;
        Stemmer stemmer = null;
        String stemmerJS = null;
        WordCollector wordCollector = new WordCollector();
        SearchInfo searchInfo = new SearchInfo();
        ArrayList<Document> saveList =
            embedCSS? new ArrayList<Document>() : null;

        for (int i = 0; i < count; ++i) {
            URL xhtmlURL = xhtmlURLs[i];

            String outFileName = URLUtil.getRawBaseName(xhtmlURL);
            File outFile = new File(outDir, URIComponent.decode(outFileName));
            info(Msg.msg("processingXHTMLFile", outFile,
                         URLUtil.toLabel(xhtmlURL)));

            Document doc = LoadDocument.load(xhtmlURL);

            if (idMap != null) {
                collectedIds.clear();
                collectIds(doc, collectedIds);

                if (collectedIds.size() > 0) {
                    String[] ids = collectedIds.toArray(StringUtil.EMPTY_LIST);
                    if (ids.length > 1) {
                        Arrays.sort(ids);
                    }

                    idMap.add(outFileName);
                    idMap.add(ids);
                }
            }

            // ---

            if (i == 0) {
                // Use the lang of the first page if specified, defaultLang
                // otherwise.
                docLang = DOMUtil.getXHTMLLang(doc.getDocumentElement());
                if (docLang == null) {
                    docLang = defaultLang;
                }

                if (useStemming) {
                    stemmer = Stemmer.create(docLang);
                    if (stemmer != null) {
                        stemmerJS = loadStemmerJS(stemmer.langCode, 
                                                  resourceList);
                        if (stemmerJS == null) {
                            stemmer = null;
                        }
                    }
                }
            }

            collectSearchInfo(doc, xhtmlURL, stemmer, wordCollector,
                              searchInfo);

            // ---

            if (!processPage(doc, pageTemplateInfo,
                             /*savePage*/ !embedCSS, outFile)) {
                return false;
            }
            
            if (embedCSS) {
                saveList.add(doc);
            }
        }

        // Save ID map ---

        if (idMap != null) {
            saveIdMap(idMap, new File(whDir, "idMap.xml"));
        }

        // Save search data ---

        info(Msg.msg("creatingSearchData"));
        String searchJS = searchInfo.toJS(stemmerJS);
        if (testMode) {
            searchJS = Util.EMACS_UTF8_DECL + searchJS;
        } else {
            searchJS = Util.compactJS(searchJS);
        }

        File outFile = new File(whDir, "search.js");
        info(Msg.msg("savingSearchData", searchInfo.wordMap.size(), 
                     searchInfo.baseNameList.size(), outFile));
        FileUtil.saveString(searchJS, outFile, "UTF-8");

        // Translate TOC to JavaScript ---

        info(Msg.msg("creatingTOC"));
        String tocJS = null;
        if (tocURL == null) {
            tocJS = simpleTOCToJS(searchInfo.titleList, 
                                  searchInfo.baseNameList);
        } else {
            tocJS = tocToJS(tocURL);
        }
        if (tocJS == null) {
            return false;
        }

        if (testMode) {
            outFile = new File(whDir, "toc.js");
            info(Msg.msg("savingTOC", outFile));
            FileUtil.saveString(Util.EMACS_UTF8_DECL + tocJS, 
                                outFile, "UTF-8");
        }

        // Translate Index to JavaScript ---
        
        assert(docLang != null);

        String indexJS = null;
        if (indexDoc != null) {
            info(Msg.msg("creatingIndex"));
            indexJS = indexToJS(indexDoc, docLang);
            if (indexJS == null) {
                return false;
            }

            if (testMode) {
                outFile = new File(whDir, "index.js");
                info(Msg.msg("savingIndex", outFile));
                FileUtil.saveString(Util.EMACS_UTF8_DECL + indexJS, 
                                    outFile, "UTF-8");
            }
        }

        // Compile and save Web Help "runtime" to _wh/ ---

        info(Msg.msg("compilingCSS"));
        String layoutMarker = "." + layout + ".";
        String[] layoutMarkers = new String[layoutList.size()];
        int k = 0;
        for (String l : layoutList) {
            layoutMarkers[k++] = "." + l + ".";
        }
        String cssRuntime = compileCSS(themeProperties,
                                       layoutMarker, layoutMarkers, 
                                       responsiveUI, resourceList);
        if (cssRuntime == null) {
            return false;
        }
        if (testMode) {
            cssRuntime = "@charset \"UTF-8\";\n\n" + cssRuntime;
        } else {
            cssRuntime = Util.compactCSS(cssRuntime);
        }

        outFile = new File(whDir, "wh.css");
        info(Msg.msg("savingCSS", outFile));
        FileUtil.saveString(cssRuntime, outFile, "UTF-8");

        // ---

        info(Msg.msg("compilingJS"));
        String uiLang = uiLanguage;
        if (uiLang != null) {
            if (uiLang.equals(BROWSER_LANGUAGE)) {
                uiLang = null;
            } else if (uiLang.equals(DOCUMENT_LANGUAGE)) {
                uiLang = docLang;
            }
        }
        UUID uuid = UUID.randomUUID();
        String storageId =
            Long.toString(uuid.getMostSignificantBits(), Character.MAX_RADIX) +
            Long.toString(uuid.getLeastSignificantBits(), Character.MAX_RADIX);
        String jsRuntime = compileJS(uiLang, storageId, tocJS, indexJS,
                                     layoutMarker, layoutMarkers,
                                     responsiveUI, resourceList);
        if (jsRuntime == null) {
            return false;
        }
        if (testMode) {
            jsRuntime = Util.EMACS_UTF8_DECL + jsRuntime;
        } else {
            jsRuntime = Util.compactJS(jsRuntime);
        }

        outFile = new File(whDir, "wh.js");
        info(Msg.msg("savingJS", outFile));
        FileUtil.saveString(jsRuntime, outFile, "UTF-8");

        // ---

        if (embedCSS) {
            for (Document doc : saveList) {
                URL saveURL = null;
                try {
                    saveURL = URLUtil.createURL(doc.getDocumentURI());
                } catch (Exception shouldNotHappen) {}
                if (saveURL == null) {
                    internalError("no save URL for page having embedded CSS");
                    return false;
                }
                
                embedCSS(doc, saveURL);

                File saveFile = URLUtil.urlToFile(saveURL);
                if (saveFile == null) {
                    internalError("no save file for page having embedded CSS");
                    return false;
                }
                
                savePage(doc, saveFile);
            }
        }
        
        return true;
    }

    private File getOutDir(URL[] xhtmlURLs) {
        File outDir = null;

        final int count = xhtmlURLs.length;
        for (int i = 0; i < count; ++i) {
            URL xhtmlURL = xhtmlURLs[i];

            File xhtmlFile = URLUtil.urlToFile(xhtmlURL);
            if (xhtmlFile != null) {
                File dir = xhtmlFile.getParentFile();
                if (outDir == null) {
                    outDir = dir;
                } else {
                    if (!outDir.equals(dir)) {
                        error(Msg.msg("multipleBaseDirs", outDir));
                        return null;
                    }
                }
            } else {
                error(Msg.msg("mustSpecifyOutDir"));
                return null;
            }
        }

        return outDir;
    }

    private String copyExtraCSS(File whDir) 
        throws IOException {
        String path = null;

        if (userCSS != null || userResources != null) {
            File userDir = new File(whDir, "user");
            if (!userDir.isDirectory()) {
                FileUtil.checkedMkdir(userDir);
            }

            if (userCSS != null) {
                String rawName = URLUtil.getRawBaseName(userCSS);
                FileUtil.copyFile(userCSS, 
                                  new File(userDir,
                                           URIComponent.decode(rawName)));
                path = "_wh/user/" + rawName;
            }

            if (userResources != null) {
                FileUtil.copyDir(userResources, 
                                 new File(userDir, userResources.getName()), 
                                 /*sameDate*/ false);
            }
        }

        return path;
    }

    private static URL listTemplateResources(URL manifestURL, String layout,
                                             List<String> layoutList,
                                             List<URL> resourceList) 
        throws IOException {
        URL pageTemplateURL = null;

        String manifest = URLUtil.loadString(manifestURL, "UTF-8");
        String[] lines = StringUtil.split(manifest, '\n');

        final String pageTemplateName = layout + ".html";
        for (String line : lines) {
            line = line.trim();

            URL url = null;
            try {
                url = URLUtil.createURL(manifestURL, line);
            } catch (MalformedURLException e) {
                throw new IOException(Msg.msg("badManifest", 
                                              URLUtil.toLabel(manifestURL), 
                                              line));
            }

            if (line.startsWith("_wh/")) {
                resourceList.add(url);
            } else {
                if (line.endsWith(".html")) {
                    layoutList.add(line.substring(0, line.length()-5));

                    if (pageTemplateName.equals(line)) {
                        pageTemplateURL = url;
                    }
                }
            }
        }

        Collections.sort(resourceList, new Comparator<URL>() {
                public int compare(URL u1, URL u2) {
                    // Compare basenames without the extension.
                    String r1 = URLUtil.getBaseName(u1);
                    String r2 = URLUtil.getBaseName(u2);
                    if (r1 == null || r2 == null) {
                        return URLUtil.COMPARATOR.compare(u1, u2);
                    } else {
                        r1 = URIComponent.setExtension(r1, null);
                        r2 = URIComponent.setExtension(r2, null);
                        return r1.compareTo(r2);
                    }
                }
            });

        return pageTemplateURL;
    }

    private Document checkEmptyIndex(URL indexURL)
        throws IOException {
        Document index = LoadDocument.load(indexURL);

        boolean isEmpty = true;

        Node node = index.getDocumentElement().getFirstChild();
        while (node != null) {
            if (node.getNodeType() == Node.ELEMENT_NODE &&
                "entry".equals(node.getLocalName())) {
                isEmpty = false;
                break;
            }

            node = node.getNextSibling();
        }

        if (isEmpty) {
            info(Msg.msg("ignoringEmptyIndex", URLUtil.toLabel(indexURL)));
            index = null;
        }

        return index;
    }

    private String loadStemmerJS(String lang, List<URL> resourceList) {
        final String baseName = "snowball_" + lang + ".js";

        URL url = null;
        for (URL resource : resourceList) {
            if (baseName.equals(URLUtil.getBaseName(resource))) {
                url = resource;
                break;
            }
        }

        String stemmerJS = null;
        if (url != null) {
            try {
                stemmerJS = URLUtil.loadString(url, "UTF-8");
                stemmerJS = stemmerJS.trim();
            } catch (IOException e) {
                error(Msg.msg("cannotLoadFile", URLUtil.toLabel(url), 
                              ThrowableUtil.reason(e)));
            }
        }

        return stemmerJS;
    }

    // -----------------------------------
    // Page template
    // -----------------------------------

    private PageTemplateInfo preparePageTemplate(URL url,
                                                 File resourceDir,
                                                 String extraCSS,
                                                 boolean[] useIndex) 
        throws IOException {
        Document doc = LoadDocument.load(url);
        
        Element html = doc.getDocumentElement();
        if (noRole) {
            removeRoleAttribute(html);
        }

        // Prepare head ---

        Element head = DOMUtil.getChildElementByLocalName(html, "head");
        if (head == null) {
            error(html, Msg.msg("missingElement", "head"));
            return null;
        }

        if (extraCSS != null) {
            Element link = doc.createElementNS(XHTML_NS_URI, "link");
            link.setAttributeNS(null, "type", "text/css");
            link.setAttributeNS(null, "rel", "stylesheet");
            link.setAttributeNS(null, "href", extraCSS);

            head.appendChild(link);
        }

        Element element = DOMUtil.getChildElementByLocalName(head, "title");
        if (element == null) {
            error(head, Msg.msg("missingElement", "title"));
            return null;
        }

        // First pass (happens after title element) ---

        Node node = element.getNextSibling();
        while (node != null) {
            if (node.getNodeType() == Node.ELEMENT_NODE) {
                Element child = (Element) node;

                String id = DOMUtil.getAttribute(child, null, "id");
                if (id != null) {
                    String attrName = null;
                    String attrValue = null;

                    if ("wh-jquery".equals(id)) {
                        attrName = "src";
                        if (jQuery != null) {
                            attrValue = jQuery;
                        }
                    }

                    if (attrName != null && attrValue != null) {
                        child.setAttributeNS(null, attrName, attrValue);
                    }
                }
            }

            node = node.getNextSibling();
        }

        // Second pass; collect metas, links and scripts after title element ---

        ArrayList<Element> includeList = new ArrayList<Element>();

        node = element.getNextSibling();
        while (node != null) {
            if (node.getNodeType() == Node.ELEMENT_NODE) {
                Element child = (Element) node;

                String id = DOMUtil.getAttribute(child, null, "id");
                if (id != null) {
                    String attrName = null;
                    String attrValue = null;

                    if ("wh-jquery".equals(id)) {
                        attrName = "src";
                        if (localJQuery) {
                            attrValue = toLocalJQuery(child, attrName,
                                                      resourceDir);
                        }
                    }

                    if (attrName != null) {
                        child.removeAttributeNS(null, "id");

                        if (attrValue != null) {
                            child.setAttributeNS(null, attrName, attrValue);
                        }
                    }
                }

                includeList.add(child);
            }

            node = node.getNextSibling();
        }

        Element[] includes = new Element[includeList.size()];
        includeList.toArray(includes);

        // Prepare body ---

        Element body = DOMUtil.getChildElementByLocalName(html, "body");
        if (body == null) {
            error(html, Msg.msg("missingElement", "body"));
            return null;
        }

        Element indexTab = DOMUtil.findElementById(body, "wh-index-tab");
        if (indexTab == null) {
            // This layout has no Index UI parts ==> Index not useful.
            useIndex[0] = false;
        }
        if (!useIndex[0]) {
            // Discard all Index UI parts if any.
            element = indexTab;
            if (element != null) {
                element.getParentNode().removeChild(element);

                element = DOMUtil.findElementById(body, "wh-index-container");
                if (element != null) {
                    element.getParentNode().removeChild(element);
                }
            }
        }

        if (!prepareHeaderFooter(body, "wh-header", userHeader)) {
            return null;
        }

        if (!prepareHeaderFooter(body, "wh-footer", userFooter)) {
            return null;
        }

        Element contentDiv = DOMUtil.findElementById(body, "wh-content");
        if (contentDiv == null) {
            error(body, Msg.msg("missingElementWithID", "div", "wh-content"));
            return null;
        }

        return new PageTemplateInfo(doc, includes, body);
    }

    private static void removeRoleAttribute(Element tree) {
        tree.removeAttributeNS(null, "role");
        
        Node node = tree.getFirstChild();
        while (node != null) {
            if (node.getNodeType() == Node.ELEMENT_NODE) {
                removeRoleAttribute((Element) node);
            }

            node = node.getNextSibling();
        }
    }
    
    private String toLocalJQuery(Element element, String attrName,
                                 File resourceDir) {
        String srcPath =  DOMUtil.getAttribute(element, null, attrName);
        String dstPath = null;
        if (srcPath != null) {
            URL srcURL = null;
            try {
                srcURL = URLUtil.createURL(srcPath);
            } catch (MalformedURLException ignored) {}

            if (srcURL != null) {
                File dstDir = new File(resourceDir, "jquery");
                String dstName = URIComponent.getRawBaseName(srcPath);
                File dstFile = new File(dstDir, URIComponent.decode(dstName));

                try {
                    copyFile(srcURL, dstFile);

                    dstPath = "_wh/jquery/" + dstName;
                } catch (Exception e) {
                    warning(Msg.msg("cannotCopyFile", srcURL, dstFile,
                                    ThrowableUtil.reason(e)));
                    // Not fatal.
                }
            }
            // Does nothing at all for a non-absolute URL.
        }

        return (dstPath == null)? srcPath : dstPath;
    }

    private void copyFile(URL srcURL, File dstFile) 
        throws IOException {
        info(Msg.msg("copyingFile", srcURL, dstFile)); 

        File dir = dstFile.getParentFile();
        if (dir != null && !dir.isDirectory()) {
            FileUtil.checkedMkdirs(dir);
        }

        InputStream in = URLUtil.openStreamNoCache(srcURL);
        try {
            FileUtil.copyFile(in, dstFile);
        } finally {
            in.close();
        }
    }

    private boolean prepareHeaderFooter(Element body, String id, 
                                        URL sourceURL) 
        throws IOException {
        Element element = DOMUtil.findElementById(body, id);
        if (sourceURL == null) {
            if (element != null) {
                element.getParentNode().removeChild(element);
            }
        } else {
            if (element == null) {
                error(body, Msg.msg("missingElementWithID", "div", id));
                return false;
            }

            Document sourceDoc;
            if ("wh-header".equals(id)) {
                if (userHeaderDoc == null) {
                    userHeaderDoc = LoadDocument.load(sourceURL);
                }
                sourceDoc = userHeaderDoc;
            } else {
                if (userFooterDoc == null) {
                    userFooterDoc = LoadDocument.load(sourceURL);
                }
                sourceDoc = userFooterDoc;
            }

            Element sourceHtml = sourceDoc.getDocumentElement();
            Element sourceBody = 
                DOMUtil.getChildElementByLocalName(sourceHtml, "body");
            if (sourceBody == null) {
                error(sourceHtml, Msg.msg("missingElement", "body"));
                return false;
            }

            DOMUtil.copyChildren(sourceBody, element, 
                                 element.getOwnerDocument());
        }

        return true;
    }

    // -----------------------------------
    // Collect IDs
    // -----------------------------------

    private void collectIds(Document doc, Set<String> ids) {
        Element html = doc.getDocumentElement();
        Element body = DOMUtil.getChildElementByLocalName(html, "body");
        if (body == null) {
            warning(html, Msg.msg("missingElement", "body"));
            return;
        }

        collectIds(body, idMapIdPrefix, ids);
    }

    private static void collectIds(Element tree, String idPrefix,
                                   Set<String> ids) {
        if (!XHTML_NS_URI.equals(tree.getNamespaceURI())) {
            return;
        }

        String id = null;
        if (idPrefix != null) {
            id = DOMUtil.getAttribute(tree, null, "id");
            if (id != null && !id.startsWith(idPrefix)) {
                id = null;
            }
        } 

        if (id == null) {
            String localName = tree.getLocalName();
            if ("h1".equals(localName) ||
                "h2".equals(localName) ||
                "h3".equals(localName) ||
                "h4".equals(localName) ||
                "h5".equals(localName) ||
                "h6".equals(localName)) {
                id = DOMUtil.getAttribute(tree, null, "id");
                if (id == null) {
                    Element firstChild = DOMUtil.getFirstChildElement(tree);
                    if (firstChild != null && 
                        DOMUtil.hasName(firstChild, XHTML_NS_URI, "a")) {
                        // Just an anchor?
                        String text;
                        if (DOMUtil.getAttribute(firstChild,
                                                 null, "href") == null && 
                            ((text = firstChild.getTextContent()) == null ||
                             text.length() == 0)) {
                            // DocBook XSL stylesheets when generating HTML4.
                            id = DOMUtil.getAttribute(firstChild, null, "name");
                            if (id == null) {
                                // DocBook XSL stylesheets when generating
                                // XHTML1.
                                id = 
                                  DOMUtil.getAttribute(firstChild, null, "id");
                            }
                        }
                    }
                }
            } else if ("section".equals(localName)) {
                // ebookc; ditac when generating XHTML5;
                // DocBook XSL stylesheets when generating XHTML5.
                id = DOMUtil.getAttribute(tree, null, "id");
            } else if ("figure".equals(localName)) {
                // Formal figure.
                Element child;
                if (((child = DOMUtil.getFirstChildElement(tree)) != null && 
                     DOMUtil.hasName(child, XHTML_NS_URI, "figcaption")) ||
                    ((child = DOMUtil.getLastChildElement(tree)) != null && 
                     DOMUtil.hasName(child, XHTML_NS_URI, "figcaption"))) {
                    id = DOMUtil.getAttribute(tree, null, "id");
                }
            } else if ("table".equals(localName)) {
                // Formal table.
                Element firstChild = DOMUtil.getFirstChildElement(tree);
                if (firstChild != null && 
                    DOMUtil.hasName(firstChild, XHTML_NS_URI, "caption")) {
                    id = DOMUtil.getAttribute(tree, null, "id");
                }
            } else if ("div".equals(localName) ||
                       "blockquote".equals(localName)) {
                // ditac when generating XHTML1: div acting as a section.
                // ebookc: formal note.
                Element firstChild = DOMUtil.getFirstChildElement(tree);
                if (firstChild != null && 
                    DOMUtil.hasName(firstChild, XHTML_NS_URI,
                                    "h1", "h2", "h3", "h4", "h5", "h6")) {
                    id = DOMUtil.getAttribute(tree, null, "id");
                }
            }
        }

        if (id != null && XMLText.isNCName(id)) {
            ids.add(id);
        }

        // ---

        Node node = tree.getFirstChild();
        while (node != null) {
            if (node.getNodeType() == Node.ELEMENT_NODE) {
                collectIds((Element) node, idPrefix, ids);
            }

            node = node.getNextSibling();
        }
    }

    private void saveIdMap(ArrayList<Object> idMap, File outFile) 
        throws IOException {
        info(Msg.msg("savingIdMap", outFile));

        OutputStreamWriter out = 
            new OutputStreamWriter(new FileOutputStream(outFile), "UTF-8");
        try {
            out.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
            out.write("<idMap");
            out.write(" title=");
            out.write(XMLText.quoteXML(idMapTitle));
            out.write(" href=");
            out.write(XMLText.quoteXML(idMapHref));
            out.write(">\n");

            final int count = idMap.size();
            for (int i = 0; i < count; i += 2) {
                String href = (String) idMap.get(i);
                String[] ids = (String[]) idMap.get(i+1);

                out.write("  <page href=");
                out.write(XMLText.quoteXML(href));
                out.write(" ids=");
                out.write(XMLText.quoteXML(StringUtil.join(' ', ids)));
                out.write("/>\n");
            }

            out.write("</idMap>\n");
            out.flush();
        } finally {
            out.close();
        }
    }

    // -----------------------------------
    // Search data
    // -----------------------------------

    private void collectSearchInfo(Document doc, URL docURL,
                                   Stemmer stemmer,
                                   WordCollector wordCollector,
                                   SearchInfo searchInfo) {
        int docIndex = searchInfo.baseNameList.size();
        // A basename should be quoted but not escaped.
        searchInfo.baseNameList.add(URLUtil.getRawBaseName(docURL));

        // title ---

        String title = null;

        Element html = doc.getDocumentElement();
        Element element = DOMUtil.getChildElementByLocalName(html, "head");
        if (element != null) {
            element = DOMUtil.getChildElementByLocalName(element, "title");
            if (element != null) {
                title = element.getTextContent();
                if (title != null) {
                    title = XMLText.collapseWhiteSpace(title);
                    if (title.length() == 0) {
                        title = null;
                    }
                }
            }
        }
        
        if (title == null) {
            warning(html, Msg.msg("missingElement", "head/title"));
            title = "???";
        }
        // A title should be escaped.
        searchInfo.titleList.add(XMLText.escapeXML(title));

        // words in body ---
        // (Do not index the title.)

        element = DOMUtil.getChildElementByLocalName(html, "body");
        if (element == null) {
            warning(html, Msg.msg("missingElement", "body"));
            return;
        }
        
        String lang = DOMUtil.getXHTMLLang(html);
        if (lang == null) {
            lang = defaultLang;
        }
        wordCollector.collect(element, lang, docIndex, stemmer, indexNumbers,
                              searchInfo.wordMap);
    }

    // -----------------------------------
    // Page processing
    // -----------------------------------

    private boolean processPage(Document doc, 
                                PageTemplateInfo pageTemplateInfo, 
                                boolean savePage, File saveFile) 
        throws IOException {
        Element html = doc.getDocumentElement();
        
        // Add a standard HTML5 "<!DOCTYPE html>" to the decorated page? ---

        boolean isHTML5 = false;
        DocumentType docType = doc.getDoctype();
        String field;
        if (docType == null ||
            (((field = docType.getPublicId()) == null || 
              field.length() == 0) &&
             ((field = docType.getSystemId()) == null || 
              field.length() == 0) &&
             ((field = docType.getInternalSubset()) == null ||
              field.length() == 0))) {
            isHTML5 = true;
        }
        if (isHTML5) {
            DocumentType newDocType = 
                doc.getImplementation().createDocumentType("html", null, null);

            if (docType == null) {
                doc.insertBefore(newDocType, html);
            } else {
                doc.replaceChild(newDocType, docType);
            }
        }

        // Process head ---

        Element head = DOMUtil.getChildElementByLocalName(html, "head");
        if (head == null) {
            error(html, Msg.msg("missingElement", "head"));
            return false;
        }

        // Replace page charset but our UTF-8 charset. ---

        Node node = head.getFirstChild();
        while (node != null) {
            Node nextNode = node.getNextSibling();

            if (node.getNodeType() == Node.ELEMENT_NODE) {
                Element child = (Element) node;

                String value = DOMUtil.getAttribute(child, null, "charset");
                if (value == null) {
                    value = DOMUtil.getAttribute(child, null, "http-equiv");
                    if (value != null &&
                        !"content-type".equalsIgnoreCase(value)) {
                        value = null;
                    }
                }

                if (value != null) {
                    head.removeChild(child);
                }
            }

            node = nextNode;
        }

        Element meta = doc.createElementNS(XHTML_NS_URI, "meta");
        if (isHTML5) {
            meta.setAttributeNS(null, "charset", "UTF-8");
        } else {
            meta.setAttributeNS(null, "http-equiv", "content-type");
            meta.setAttributeNS(null, "content", "text/html; charset=UTF-8");
        }
        head.insertBefore(meta, head.getFirstChild());

        // Insert whc scripts, stylesheets, meta. ---
        //
        // - whc scripts are inserted before first page script if any;
        //  appended to the head otherwise.
        //  (Note that as of jQuery v3, document.ready is 
        //   called asynchronously.)
        // 
        // - whc stylesheets and meta are appended to the head.

        Element firstUserScript = null;

        node = meta.getNextSibling();
        while (node != null) {
            if (node.getNodeType() == Node.ELEMENT_NODE) {
                Element child = (Element) node;

                if ("script".equals(child.getLocalName())) {
                    firstUserScript = child;
                    break;
                }
            }

            node = node.getNextSibling();
        }

        for (Element include : pageTemplateInfo.includes) {
            Node includeCopy = doc.importNode(include, /*deep*/ true);

            if ("script".equals(includeCopy.getLocalName()) && 
                firstUserScript != null) {
                head.insertBefore(includeCopy, firstUserScript);
            } else {
                head.appendChild(includeCopy);
            }
        }

        // Process body ---

        Element body = DOMUtil.getChildElementByLocalName(html, "body");
        if (body == null) {
            error(html, Msg.msg("missingElement", "body"));
            return false;
        }

        // This is needed to copy the watermark set by ditac.
        String bodyStyle = DOMUtil.getAttribute(body, null, "style");

        html.removeChild(body);

        Element newBody = (Element) 
            doc.importNode(pageTemplateInfo.body, /*deep*/ true);
        html.appendChild(newBody);

        Element contentDiv = DOMUtil.findElementById(newBody, "wh-content");
        assert(contentDiv != null);
        DOMUtil.copyChildren(body, contentDiv, doc);

        if (bodyStyle != null) {
            // This is needed to copy the watermark set by ditac.

            List<String> styleList = splitStyleProps(bodyStyle);
            Iterator<String> iter = styleList.iterator();
            while (iter.hasNext()) {
                String style = iter.next();

                if (!style.startsWith("background")) {
                    iter.remove();
                }
            }

            if (styleList.size() > 0) {
                contentDiv.setAttributeNS(null, "style",
                                          joinStyleProps(styleList));
            }
        }

        if (savePage) {
            savePage(doc, saveFile);
        } else {
            // Will be saved later.
            doc.setDocumentURI(FileUtil.fileToURL(saveFile).toExternalForm());
        }

        return true;
    }

    @SuppressWarnings("fallthrough")
    private static List<String> splitStyleProps(String styles) {
        ArrayList<String> list = new ArrayList<String>();
        char quote = 0;
        StringBuilder buffer = null;

        final int length = styles.length();
        loop: for (int i = 0; i < length; ++i) {
            char c = styles.charAt(i);

            switch (c) {
            case ';':
                if (quote != 0) {
                    // ';' inside quoted string.
                    buffer.append(c);
                } else {
                    // End of style.
                    if (buffer != null) {
                        String style = buffer.toString().trim();
                        if (style.length() > 0) {
                            list.add(style);
                        }
                        buffer = null;
                    }
                }
                break;

            case '\"': case '\'':
                if (quote != 0) {
                    if (c == quote) {
                        int last = buffer.length()-1;
                        if (last < 0 || buffer.charAt(last) != '\\') {
                            // End of quoted string.
                            quote = 0;
                        }
                        // Otherwise, escaped quote.
                    }
                    // Otherwise, the other quote char (e.g. '\'' inside
                    // a '"' quoted string).
                } else {
                    // Begin quoted string.
                    quote = c;
                }
                /*FALLTHROUGH*/

            default:
                if (buffer == null) {
                    buffer = new StringBuilder();
                }
                buffer.append(c);
                break;
            }
        }

        if (buffer != null) {
            String style = buffer.toString().trim();
            if (style.length() > 0) {
                list.add(style);
            }
        }

        return list;
    }

    private static String joinStyleProps(List<String> styles) {
        StringBuilder buffer = new StringBuilder();

        boolean first = true;
        for (String style : styles) {
            if (first) {
                first = false;
            } else {
                buffer.append(' ');
            }

            buffer.append(style);
            buffer.append(';');
        }

        return buffer.toString();
    }

    private static void savePage(Document doc, File outFile) 
        throws IOException {
        File tmpFile = new File(outFile.getParentFile(), 
                                outFile.getName() + ".WHC");
        SaveDocument.saveXHTML(doc, tmpFile);

        if (outFile.exists()) {
            FileUtil.deleteFileOrDir(outFile);
        }
        FileUtil.checkedRename(tmpFile, outFile);
    }

    // -----------------------------------
    // embedCSS
    // -----------------------------------

    private boolean embedCSS(Document doc, URL docURL) {
        Element html = doc.getDocumentElement();
        Element head = DOMUtil.getChildElementByLocalName(html, "head");
        if (head == null) {
            error(html, Msg.msg("missingElement", "head"));
            return false;
        }

        Node node = head.getFirstChild();
        while (node != null) {
            Node nextNode = node.getNextSibling();

            if (node.getNodeType() == Node.ELEMENT_NODE) {
                Element child = (Element) node;

                if (child.getLocalName().equals("link")) {
                    String css = loadCSS(child, docURL);
                    if (css != null) {
                        Element style =
                            doc.createElementNS(XHTML_NS_URI, "style");
                        style.setAttributeNS(null, "type", "text/css");
                        style.appendChild(doc.createTextNode(css));
                        head.replaceChild(style, child);
                    }
                }
            }

            node = nextNode;
        }

        return true;
    }
    
    private String loadCSS(Element link, URL baseURL) {
        String href = DOMUtil.getAttribute(link, null, "href");
        if (href == null) {
            return null;
        }
        
        if (!"text/css".equalsIgnoreCase(DOMUtil.getAttribute(link,
                                                              null, "type")) &&
            !href.endsWith(".css")) {
            // Simplification.
            return null;
        }
        
        String rel = DOMUtil.getAttribute(link, null, "rel");
        if (rel == null || (rel = rel.trim().toLowerCase()).length() == 0 ||
            // Simplification.
            rel.indexOf("stylesheet") < 0 || rel.indexOf("alternate") >= 0) {
            return null;
        }

        // Simplification: <base> not taken into account here.
        URL url = null;
        try {
            url = URLUtil.createURL(baseURL, href);
        } catch (MalformedURLException ignored) {}
        if (url == null) {
            error(link, Msg.msg("invalidURLOrFile", href));
            return null;
        }

        // IMPORTANT LIMITATION: embedCSS cannot work unless the CSS has been
        // copied to the output directory BEFORE whc is run to decorate the
        // HTML pages contained in this output directory.
        //
        // This is always the case for "_wh/wh.css" (see compile()) but is it
        // the case for the user's CSS stylesheets?
        
        String css = null;
        try {
            css = LoadText.loadText(url, "UTF-8", /*out_encoding*/ null, 
                                    LoadText.CSS_CHARSET_DETECTOR);
        } catch (IOException e) {
            error(link,Msg.msg("cannotLoadFile", url, ThrowableUtil.reason(e)));
            return null;
        }
        
        css = Util.compactCSS(css);

        // No need to process the url(XXX) contained in embedded stylesheet.
        // The stylesheet is supposed to have already been copied to output
        // directory and to be usable from this location.
        
        return css;
    }
    
    // -----------------------------------
    // TOC
    // -----------------------------------

    private String simpleTOCToJS(List<String> titleList, 
                                 List<String> baseNameList) {
        StringBuilder buffer = new StringBuilder();
        buffer.append("var toc_entries = [\n");
        
        int last = titleList.size() - 1;
        for (int i = 0; i <= last; ++i) {
            buffer.append('[');

            // Already escaped.
            Util.appendJSString(titleList.get(i), buffer);
            buffer.append(',');
            // Already quoted but not escaped.
            Util.appendJSString(baseNameList.get(i), buffer);

            if (i == last) {
                buffer.append(",null]\n");
            } else {
                buffer.append(",null],\n");
            }
        }

        buffer.append("];\n");

        appendCollapseTOC(buffer);

        return buffer.toString();
    }

    private void appendCollapseTOC(StringBuilder buffer) {
        buffer.append("var toc_initiallyCollapsed = ");
        buffer.append(Boolean.toString(collapseTOC));
        buffer.append(";\n");
    }

    private String tocToJS(URL tocURL) 
        throws IOException {
        Document toc = LoadDocument.load(tocURL);

        // ---

        StringBuilder buffer = new StringBuilder();
        buffer.append("var toc_entries = [\n");

        if (!tocEntriesToJS(toc.getDocumentElement(), buffer)) {
            return null;
        }

        buffer.append("];\n");

        appendCollapseTOC(buffer);

        return buffer.toString();
    }

    private boolean tocEntriesToJS(Element container, StringBuilder buffer) {
        boolean done = true;

        boolean first = true;
        Node node = container.getFirstChild();
        while (node != null) {
            if (node.getNodeType() == Node.ELEMENT_NODE &&
                "entry".equals(node.getLocalName())) {
                if (first) {
                    first = false;
                } else {
                    buffer.append(",\n");
                }

                if (!tocEntryToJS((Element) node, buffer)) {
                    done = false;
                    break;
                }
            }

            node = node.getNextSibling();
        }

        return done;
    }

    private boolean tocEntryToJS(Element entry, StringBuilder buffer) {
        String href = DOMUtil.getAttribute(entry, null, "href");
        if (href != null) {
            href = XMLText.escapeXML(href);
        }

        String title = null;
        boolean hasEntries = false;

        Node node = entry.getFirstChild();
        while (node != null) {
            if (node.getNodeType() == Node.ELEMENT_NODE) {
                String localName = node.getLocalName();

                if ("title".equals(localName)) {
                    title = SaveDocument.xhtmlToString(node);
                } else if ("entry".equals(localName)) {
                    hasEntries = true;
                    break;
                }
            }

            node = node.getNextSibling();
        }

        if (title == null) {
            error(entry, Msg.msg("missingElement", "title"));
            return false;
        }

        boolean done = true;

        buffer.append('[');

        Util.appendJSString(title, buffer);
        if (href == null) {
            buffer.append(",null");
        } else {
            buffer.append(',');
            Util.appendJSString(href, buffer);
        }

        if (hasEntries) {
            buffer.append(",[\n");
            done = tocEntriesToJS(entry, buffer);
            buffer.append("\n]");
        } else {
            buffer.append(",null");
        }

        buffer.append(']');

        return done;
    }

    // -----------------------------------
    // Index
    // -----------------------------------

    private String indexToJS(Document indexDoc, String docLang) 
        throws IOException {
        IndexEntry[] entries = loadIndex(indexDoc, docLang);
        if (entries == null) {
            return null;
        }

        // ---

        StringBuilder buffer = new StringBuilder();
        buffer.append("var index_entries = [\n");

        if (!indexEntriesToJS(entries, buffer)) {
            return null;
        }

        buffer.append("\n];\n");

        return buffer.toString();
    }

    // ---------
    // loadIndex
    // ---------

    private IndexEntry[] loadIndex(Document indexDoc, String docLang) 
        throws IOException {
        Element index = indexDoc.getDocumentElement();

        String termSepar = index.getAttributeNS(null, "termSeparator");
        if (termSepar == null || termSepar.length() == 0) {
            termSepar = ", ";
        }

        String lang = DOMUtil.getAttribute(index, XML_NS_URI, "lang");
        if (lang == null) {
            lang = docLang;
        }
        assert(lang != null);
        Locale locale = LocaleUtil.getLocale(lang);

        boolean mergeAndSort = 
            DOMUtil.getBooleanAttribute(index, null, "mergeAndSort", false);

        boolean caseSensitive = 
            DOMUtil.getBooleanAttribute(index, null, "caseSensitive", false);

        ArrayList<IndexEntry> entryList = new ArrayList<IndexEntry>();
        // entryList may be very long.
        HashMap<String,IndexEntry> termToTopEntry = 
            new HashMap<String,IndexEntry>();

        Node node = index.getFirstChild();
        while (node != null) {
            if (node.getNodeType() == Node.ELEMENT_NODE &&
                "entry".equals(node.getLocalName())) {
                IndexEntry entry;
                try {
                    entry = IndexEntry.parse((Element) node);
                } catch (IllegalArgumentException e) {
                    String reason = e.getMessage();
                    if (reason == null || reason.length() == 0) {
                        reason = Msg.msg("invalidIndexEntry");
                    } else {
                        reason = Msg.msg("invalidIndexEntry2", reason);
                    }
                    error((Element) node, reason);
                    return null;
                }
                
                if (mergeAndSort) {
                    String term = entry.term.text;
                    if (!caseSensitive) {
                        term = term.toLowerCase(locale);
                    }

                    IndexEntry topEntry = termToTopEntry.get(term);
                    if (topEntry == null) {
                        termToTopEntry.put(term, entry);
                        entryList.add(entry);
                    } else {
                        topEntry.merge(entry, caseSensitive, locale);
                    }
                } else {
                    entryList.add(entry);
                }
            }

            node = node.getNextSibling();
        }

        // ---

        IndexEntry[] entries = new IndexEntry[entryList.size()];
        entryList.toArray(entries);
        entryList = null;

        if (!fixUpRefs(entries, termSepar)) {
            return null;
        }

        // ---

        if (mergeAndSort) {
            final Collator termCollator = createTermCollator(locale);

            Comparator<IndexEntry> entryCompare = new Comparator<IndexEntry>() {
                public int compare(IndexEntry entry1, IndexEntry entry2) {
                    String sa1 = entry1.term.sortAs;
                    String t1 = (sa1 == null)? entry1.term.text : sa1;

                    String sa2 = entry2.term.sortAs;
                    String t2 = (sa2 == null)? entry2.term.text : sa2;

                    int delta = termCollator.compare(t1, t2);
                    if (delta != 0) {
                        return delta;
                    } else {
                        return termCollator.compare(entry1.term.text, 
                                                    entry2.term.text);
                    }
                }
            };

            Comparator<IndexRef> refCompare = new Comparator<IndexRef>() {
                public int compare(IndexRef ref1, IndexRef ref2) {
                    return termCollator.compare(ref1.termText, ref2.termText);
                }
            };

            Arrays.sort(entries, entryCompare);

            for (IndexEntry entry : entries) {
                sortEntry(entry, refCompare, entryCompare);
            }
        }

        return entries;
    }
    
    private boolean fixUpRefs(IndexEntry[] entries, String termSepar) {
        Stack<IndexTerm> termStack = new Stack<IndexTerm>();
        HashMap<String,IndexTerm> idToTerm = new HashMap<String,IndexTerm>();

        for (IndexEntry entry : entries) {
            while (!termStack.empty()) {
                // Should not happen.
                termStack.pop();
            }

            mapIdToTerm(entry, termSepar, termStack, idToTerm);
        }

        for (IndexEntry entry : entries) {
            if (!fixUpRefs(entry, idToTerm)) {
                return false;
            }
        }

        return true;
    }

    private static void mapIdToTerm(IndexEntry entry, 
                                    String termSepar, 
                                    Stack<IndexTerm> termStack, 
                                    Map<String,IndexTerm> idToTerm) {
        IndexTerm term = entry.term;
        if (!termStack.empty()) {
            term = termStack.peek().append(term, termSepar);
        }

        if (entry.id != null) {
            idToTerm.put(entry.id, term);
        }

        IndexEntry[] subEntries = entry.entryList;
        if (subEntries != null) {
            termStack.push(term);

            for (IndexEntry subEntry : subEntries) {
                mapIdToTerm(subEntry, termSepar, termStack, idToTerm);
            }

            termStack.pop();
        }
    }

    private boolean fixUpRefs(IndexEntry entry, 
                              Map<String,IndexTerm> idToTerm) {
        IndexRef[] refs = entry.seeList;
        if (refs != null) {
            for (IndexRef ref : refs) {
                if (!fixUpRef("see", ref, idToTerm)) {
                    return false;
                }
            }
        }

        refs = entry.seeAlsoList;
        if (refs != null) {
            for (IndexRef ref : refs) {
                if (!fixUpRef("seeAlso", ref, idToTerm)) {
                    return false;
                }
            }
        }

        IndexEntry[] subEntries = entry.entryList;
        if (subEntries != null) {
            for (IndexEntry subEntry : subEntries) {
                fixUpRefs(subEntry, idToTerm);
            }
        }

        return true;
    }

    private boolean fixUpRef(String refType, IndexRef ref, 
                             Map<String,IndexTerm> idToTerm) {
        if (ref.termText == null) {
            if (ref.ref == null) {
                error(Msg.msg("voidElement", refType));
                return false;
            }

            IndexTerm term = idToTerm.get(ref.ref);
            if (term == null) {
                error(Msg.msg("missingElementWithID", "entry", ref.ref));
                return false;
            }

            ref.termHTML = term.html;
            ref.termText = term.text;
        }

        return true;
    }

    private static Collator createTermCollator(Locale locale) {
        Collator collator = Collator.getInstance(locale);
        collator.setDecomposition(Collator.FULL_DECOMPOSITION);
        collator.setStrength(Collator.IDENTICAL);
        return collator;
    }

    private static void sortEntry(IndexEntry entry, 
                                  Comparator<IndexRef> refCompare,
                                  Comparator<IndexEntry> entryCompare) {
        IndexRef[] refs = entry.seeList;
        if (refs != null && refs.length > 1) {
            Arrays.sort(refs, refCompare);
        }

        refs = entry.seeAlsoList;
        if (refs != null && refs.length > 1) {
            Arrays.sort(refs, refCompare);
        }

        IndexEntry[] subEntries = entry.entryList;
        if (subEntries != null) {
            if (subEntries.length > 1) {
                Arrays.sort(subEntries, entryCompare);
            }

            for (IndexEntry subEntry : subEntries) {
                sortEntry(subEntry, refCompare, entryCompare);
            }
        }
    }

    // ----------------
    // indexEntriesToJS
    // ----------------

    private boolean indexEntriesToJS(IndexEntry[] entries, 
                                     StringBuilder buffer) {
        boolean first = true;
        for (IndexEntry entry : entries) {
            if (first) {
                first = false;
            } else {
                buffer.append(",\n");
            }

            if (!indexEntryToJS(entry, buffer)) {
                return false;
            }
        }

        return true;
    }

    private boolean indexEntryToJS(IndexEntry entry, StringBuilder buffer) {
        IndexRef[] seeList = entry.seeList;
        IndexAnchor[] anchorList = entry.anchorList;
        IndexRef[] seeAlsoList = entry.seeAlsoList;
        IndexEntry[] subEntries = entry.entryList;

        if (seeList == null && anchorList == null && subEntries == null) {
            error(Msg.msg("uselessIndexEntry", entry.term.text));
            return false;
        }

        buffer.append('{');

        buffer.append("term: ");
        Util.appendJSString(entry.term.html, buffer);

        int count;
        if (seeList != null && (count = seeList.length) > 0) {
            buffer.append(",\nsee: [");
            for (int i = 0; i < count; ++i) {
                IndexRef ref = seeList[i];

                if (i > 0) {
                    buffer.append(',');
                }
                Util.appendJSString(ref.termHTML, buffer);
            }
            buffer.append(']');
        } else {
            if (anchorList != null && (count = anchorList.length) > 0) {
                buffer.append(",\nanchor: [");
                for (int i = 0; i < count; ++i) {
                    IndexAnchor anchor = anchorList[i];
                    String href = XMLText.escapeXML(anchor.href);
                    String href2 = anchor.href2;

                    if (i > 0) {
                        buffer.append(',');
                    }
                    Util.appendJSString(href, buffer);

                    if (href2 == null) {
                        buffer.append(",null");
                    } else {
                        buffer.append(',');
                        href2 = XMLText.escapeXML(href2);
                        Util.appendJSString(href2, buffer);
                    }
                }
                buffer.append(']');
            }

            if (seeAlsoList != null && (count = seeAlsoList.length) > 0) {
                buffer.append(",\nseeAlso: [");
                for (int i = 0; i < count; ++i) {
                    IndexRef ref = seeAlsoList[i];

                    if (i > 0) {
                        buffer.append(',');
                    }
                    Util.appendJSString(ref.termHTML, buffer);
                }
                buffer.append(']');
            }

            if (subEntries != null && subEntries.length > 0) {
                buffer.append(",\nentry: [\n");

                if (!indexEntriesToJS(subEntries, buffer)) {
                    return false;
                }

                buffer.append("\n]");
            }
        }

        buffer.append('}');

        return true;
    }

    // -----------------------------------
    // Compile Web Help runtime
    // -----------------------------------

    private static String[] listSkippedLayouts(String layout, 
                                               List<String> layoutList) {
        String[] skippedLayouts = StringList.EMPTY_LIST;

        for (String name : layoutList) {
            if (!layout.equals(name)) {
                skippedLayouts = StringList.append(skippedLayouts, 
                                                   "." + name + ".");
            }
        }

        return skippedLayouts;
    }

    @SuppressWarnings("unchecked")
    private String compileCSS(HashMap<String,String> themeProperties,
                              String layoutMarker, String[] layoutMarkers,
                              boolean responsive, List<URL> resourceList) 
        throws IOException {
        StringBuilder buffer = new StringBuilder();

        HashMap<String,String> cssVars = 
            (HashMap<String,String>) themeProperties.clone();

        // IE11 supports calc() but not var().
        // Example (ending with a CSS comment):
        // --tab-bg: var(--panel-bg); /*Normal background*/

        Matcher matcher = Pattern.compile("^(--[^:]+):([^;]+);").matcher("");

        for (URL resource : resourceList) {
            String baseName = URLUtil.getBaseName(resource);

            int rank = parseRank(baseName, layoutMarker, layoutMarkers,
                                 responsive, ".css");
            if (rank < 0) {
                // Skip.
                continue;
            }

            String source = URLUtil.loadString(resource, "UTF-8");

            buffer.append("/* ");
            buffer.append(baseName);
            buffer.append(" ======================================== */\n\n");

            final String[] lines = StringUtil.split(source, '\n');
            final int lineCount = lines.length;
            for (int i = 0; i < lineCount; ++i) {
                String line = lines[i];

                matcher.reset(line.trim());
                if (matcher.find()) {
                    String name = matcher.group(1).trim();
                    String value = matcher.group(2).trim();
                    value = substituteCSSVars(value, cssVars, resource, i);
                    if (value == null) {
                        return null;
                    }

                    cssVars.putIfAbsent(name, value);
                } else {
                    line = substituteCSSVars(line, cssVars, resource, i);
                    if (line == null) {
                        return null;
                    }

                    buffer.append(line);
                    buffer.append('\n');
                }
            }
        }

        return buffer.toString();
    }

    private static int parseRank(String baseName,
                                 String layoutMarker, String[] layoutMarkers, 
                                 boolean responsive, String suffix) {
        if (!baseName.endsWith(suffix)) {
            // Skip.
            return -1;
        }

        if (!responsive && baseName.lastIndexOf(".responsive.") >= 0) {
            // Skip.
            return -1;
        }

        boolean layoutSpecific = false;
        for (String marker : layoutMarkers) {
            if (baseName.lastIndexOf(marker) >= 0) {
                layoutSpecific = true;
                break;
            }
        }
        if (layoutSpecific && baseName.lastIndexOf(layoutMarker) < 0) {
            // Skip.
            return -1;
        }
        
        // Ignore all resources having a name which does not start with one or
        // more digits. Example: "snowball_en.js".
        int digitCount = 0;
        final int count = baseName.length();
        for (int i = 0; i < count; ++i) {
            char c = baseName.charAt(i);
            if (c < '0' || c > '9') {
                digitCount = i;
                break;
            }
        }
        if (digitCount == 0) {
            return -1;
        } 

        try {
            return Integer.parseInt(baseName.substring(0, digitCount));
        } catch (NumberFormatException cannotHappen) {
            return -1;
        }
    }

    private String substituteCSSVars(String text, Map<String,String> cssVars,
                                     URL resource, int lineIndex) {
        StringBuilder buffer = null;
        int from = 0;

        for (;;) {
            int start = text.indexOf("var(", from);
            if (start < 0) {
                break;
            }

            int end = text.indexOf(')', start+4);
            if (end < 0) {
                break;
            }

            String name = text.substring(start+4, end).trim();
            String value;
            if (name.length() == 0 || 
                (value = cssVars.get(name)) == null) {
                error(Msg.msg("unknownCSSVar", name, 
                              URLUtil.toLabel(resource), 
                              Integer.toString(1+lineIndex)));
                return null;
            }

            if (buffer == null) {
                buffer = new StringBuilder();
            }
            if (start > from) {
                buffer.append(text.substring(from, start));
            }
            buffer.append(value);

            from = end+1;
        }

        if (buffer == null) {
            return text;
        } else {
            buffer.append(text.substring(from));
            return buffer.toString();
        }
    }

    private String compileJS(String uiLang, String storageId, 
                             String tocJS, String indexJS,
                             String layoutMarker, String[] layoutMarkers,
                             boolean responsive, List<URL> resourceList) 
        throws IOException {
        StringBuilder buffer = new StringBuilder();

        ArrayList<String> publicFunctions = new ArrayList<String>();
        int wrapped = 0;

        for (URL resource : resourceList) {
            String baseName = URLUtil.getBaseName(resource);

            int rank = parseRank(baseName, layoutMarker, layoutMarkers,
                                 responsive, ".js");
            if (rank < 0) {
                // Skip.
                continue;
            }
            String rankPrefix = Integer.toString(rank);

            if (rank >= 20 && rank < 80) {
                if (wrapped == 0) {
                    wrapped = 1;
                    buffer.append("var wh = (function () {\n\n");
                }
            } else if (rank >= 80) {
                if (wrapped == 1) {
                    wrapped = 2;
                    endWrapJS(publicFunctions, buffer);
                }
            }

            String source;
            String sourceName;
            boolean verbatim = false;
            if (baseName.equals(rankPrefix + "toc.js")) {
                source = tocJS;
                sourceName = "toc.js";
                verbatim = true;
            } else if (baseName.equals(rankPrefix + "index.js")) {
                source = indexJS;
                sourceName = "index.js";
                verbatim = true;
            } else {
                source = URLUtil.loadString(resource, "UTF-8");
                sourceName = baseName;
            }

            if (source != null) {
                buffer.append("/* ");
                buffer.append(sourceName);
                buffer.append(" ============================== */\n\n");

                if (verbatim) {
                    buffer.append(source);
                } else {
                    final String[] lines = StringUtil.split(source, '\n');
                    for (String line : lines) {
                        String functionName = getPublicFunctionName(line);
                        if (functionName != null) {
                            publicFunctions.add(functionName);
                        } else {
                            if (line.indexOf("var storageId =") >= 0) {
                                line = "var storageId = \"" + storageId + "\";";
                            } else if (uiLang != null &&
                                       line.indexOf(
                                         "var preferredUserLanguage =") >= 0) {
                                line = "var preferredUserLanguage = \"" + 
                                    uiLang + "\";";
                            }
                        }
                        
                        buffer.append(line);
                        buffer.append('\n');
                    }
                }

                buffer.append('\n');
            }
        }

        if (wrapped == 1) {
            wrapped = 2;
            endWrapJS(publicFunctions, buffer);
        }

        return buffer.toString();
    }

    private static String getPublicFunctionName(String line) {
        final String decl = "/*public*/ function ";
        int start = line.indexOf(decl);
        if (start < 0) {
            return null;
        }
        start += decl.length();

        int end = line.indexOf('(', start);
        if (end <= start) {
            return null;
        }

        String name = line.substring(start, end).trim();
        return (name.length() == 0)? null : name;
    }

    private static void endWrapJS(List<String> publicFunctions, 
                                  StringBuilder buffer) {
        buffer.append("return {\n");
        
        for (String name : publicFunctions) {
            buffer.append("    ");
            buffer.append(name);
            buffer.append(": ");
            buffer.append(name);
            buffer.append(",\n");
        }

        buffer.append("}\n\n})();\n\n");
    }

    // -----------------------------------
    // Messages
    // -----------------------------------

    private void internalError(String message) {
        message = Msg.msg("internalError") + " " + message;
        if (console == null) {
            System.err.println(message);
        } else {
            console.showMessage(message, Console.MessageType.ERROR);
        }
    }

    private void error(String message) {
        if (console == null) {
            System.err.println(Msg.msg("error") + " " + message);
        } else {
            console.showMessage(message, Console.MessageType.ERROR);
        }
    }

    private void warning(String message) {
        if (console == null) {
            System.err.println(Msg.msg("warning") + " " + message);
        } else {
            console.showMessage(message, Console.MessageType.WARNING);
        }
    }

    private void info(String message) {
        if (verbose) {
            if (console == null) {
                System.out.println(message);
            } else {
                console.showMessage(message, Console.MessageType.INFO);
            }
        }
    }

    private void error(Element element, String message) {
        error(prependElementLocation(element, message));
    }

    private void warning(Element element, String message) {
        warning(prependElementLocation(element, message));
    }

    private static String prependElementLocation(Element element,
                                                 String message) {
        if (element != null) {
            NodeLocation location = 
                (NodeLocation) element.getUserData(NodeLocation.USER_DATA_KEY);
            if (location == null) {
                location = NodeLocation.UNKNOWN_LOCATION;
            }

            StringBuilder buffer = new StringBuilder();
            location.toString(buffer);
            buffer.append(": ");
            buffer.append(message);

            message = buffer.toString();
        }
        return message;
    }

    // -----------------------------------------------------------------------
    // Main
    // -----------------------------------------------------------------------

    public static void main(String[] args) {
        boolean verbose = false;
        URL tocURL = null;
        URL indexURL = null;
        ArrayList<String> parameterList = new ArrayList<String>();

        int i;
        String arg;
        int argCount = args.length;
        for (i = 0; i < argCount; ++i) {
            arg = args[i];

            if ("-v".equals(arg)) {
                verbose = true;
            } else if ("-toc".equals(arg)) {
                if (i+1 >= argCount) {
                    usage(null);
                }

                tocURL = URLUtil.urlOrFile(args[++i]);
                if (tocURL == null) {
                    usage(Msg.msg("invalidURLOrFile", args[i]));
                }
            } else if ("-index".equals(arg)) {
                if (i+1 >= argCount) {
                    usage(null);
                }

                indexURL = URLUtil.urlOrFile(args[++i]);
                if (indexURL == null) {
                    usage(Msg.msg("invalidURLOrFile", args[i]));
                }
            } else if ("-p".equals(arg)) {
                if (i+2 >= argCount) {
                    usage(null);
                }

                parameterList.add(args[++i]);
                parameterList.add(args[++i]);
            } else if ("-template".equals(arg)) {
                // Not documented. Only used to test whc.
                if (i+1 >= argCount) {
                    usage(null);
                }

                URL manifestURL = URLUtil.urlOrFile(args[++i]);
                if (manifestURL == null) {
                    usage(Msg.msg("invalidURLOrFile", args[i]));
                }
                Compiler.setTemplateManifestURL(manifestURL);
            } else {
                if (arg.startsWith("-") && !"-".equals(arg)) {
                    usage(null);
                }
                break;
            }
        }

        if (i + 1 >= argCount) {
            usage(null);
        }

        File outDir = null;
        arg = args[i++];
        if (!"-".equals(arg)) {
            outDir = new File(arg);
            if (!outDir.isDirectory()) {
                usage(Msg.msg("notADir", arg));
            }
        }

        URL[] xhtmlURLs = new URL[argCount - i];
        int j = 0;
        for (; i < argCount; ++i) {
            URL xhtmlURL = URLUtil.urlOrFile(args[i]);
            if (xhtmlURL == null) {
                usage(Msg.msg("invalidURLOrFile", args[i]));
            }

            xhtmlURLs[j++] = xhtmlURL;
        }

        Compiler compiler = new Compiler(null);
        compiler.setVerbose(verbose);
        if (parameterList.size() > 0) {
            if (!compiler.parseParameters(
                    parameterList.toArray(StringUtil.EMPTY_LIST))) {
                System.exit(2);
            }
        }

        String error = null;

        try {
            if (!compiler.compile(xhtmlURLs, tocURL, indexURL, outDir)) {
                error = "";
            }
        } catch (Exception e) {
            error = ThrowableUtil.reason(e);
        }

        if (error != null) {
            if (error.length() > 0) {
                System.err.println(Msg.msg("cannotCompile2", error));
            } else {
                System.err.println(Msg.msg("cannotCompile"));
            }
            System.exit(3);
        }
    }

    private static void usage(String message) {
        if (message != null) {
            System.err.println(Msg.msg("error") + " " + message);
        }
        System.err.println(Msg.msg("usage"));
        System.exit(1);
    }
}
