/*
 * Copyright (c) 2017-2019 XMLmind Software. All rights reserved.
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind Web Help Compiler project.
 * For conditions of distribution and use, see the accompanying LEGAL.txt file.
 */
package com.xmlmind.whc;

import static javax.xml.XMLConstants.XML_NS_URI;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.DOMConfiguration;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.UserDataHandler;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.Document;
import org.w3c.dom.ls.DOMImplementationLS;
import com.xmlmind.util.ThrowableUtil;
import com.xmlmind.util.ObjectUtil;

/*package*/ final class DOMUtil {
    private DOMUtil() {}

    // -----------------------------------------------------------------------

    public static Document newDocument() {
        DOMImplementation impl = getDOMImplementation();
        return impl.createDocument(null, null, null);
    }

    /*package*/ static DOMImplementation getDOMImplementation() {
        DOMImplementation impl;
        try {
            DocumentBuilder builder =
                DocumentBuilderFactory.newInstance().newDocumentBuilder();
            impl = builder.getDOMImplementation();

            // Check whether it actually implements DOM3 Load & Save.
            DOMImplementationLS implLS = (DOMImplementationLS) impl;
        } catch (Exception shouldNotHappen) {
            throw new RuntimeException(
                Msg.msg("cannotGetDOMImplementationLS",
                        ThrowableUtil.reason(shouldNotHappen)));
        }
        return impl;
    }

    // -----------------------------------------------------------------------

    public static final class CopyUserData implements UserDataHandler {
        public void handle(short operation, String key, Object data,
                           Node src, Node dst) {
            if (data != null) {
                switch (operation) {
                case UserDataHandler.NODE_IMPORTED:
                case UserDataHandler.NODE_CLONED:
                    dst.setUserData(key, data, this);
                    break;
                }
            }
        }
    }
    public static final CopyUserData COPY_USER_DATA = new CopyUserData();

    // -----------------------------------------------------------------------

    public static boolean hasName(Node node,
                                  String namespace, String... localNames) {
        String localName2 = node.getLocalName();
        if (localName2 == null ||
            !ObjectUtil.equals(namespace, node.getNamespaceURI())) {
            return false;
        }

        for (String localName : localNames) {
            if (localName.equals(localName2)) {
                return true;
            }
        }

        return false;
    }

    // -----------------------------------------------------------------------

    public static void copyChildren(Element from, Element to, Document doc) {
        Node child = from.getFirstChild();
        while (child != null) {
            to.appendChild(doc.importNode(child, /*deep*/ true));

            child = child.getNextSibling();
        }
    }

    // -----------------------------------------------------------------------

    public static Element getChildElementByLocalName(Node node,
                                                     String localName) {
        Node child = node.getFirstChild();
        while (child != null) {
            if (child.getNodeType() == Node.ELEMENT_NODE) {
                Element childElement = (Element) child;

                if (childElement.getLocalName().equals(localName)) {
                    return childElement;
                }
            }

            child = child.getNextSibling();
        }

        return null;
    }

    // -----------------------------------------------------------------------

    public static Element findElementById(Node node, String id) {
        Node child = node.getFirstChild();
        while (child != null) {
            if (child.getNodeType() == Node.ELEMENT_NODE) {
                Element childElement = (Element) child;

                String value = childElement.getAttributeNS(null, "id");
                if (value != null && value.length() > 0) {
                    value = value.trim();
                    if (value.equals(id)) {
                        return childElement;
                    }
                }

                Element found = findElementById(childElement, id);
                if (found != null) {
                    return found;
                }
            }

            child = child.getNextSibling();
        }

        return null;
    }

    // -----------------------------------------------------------------------

    public static Element getFirstChildElement(Node node) {
        Node child = node.getFirstChild();
        while (child != null) {
            if (child.getNodeType() == Node.ELEMENT_NODE) {
                return (Element) child;
            }
            child = child.getNextSibling();
        }

        return null;
    }

    public static Element getLastChildElement(Node node) {
        Node child = node.getLastChild();
        while (child != null) {
            if (child.getNodeType() == Node.ELEMENT_NODE) {
                return (Element) child;
            }
            child = child.getPreviousSibling();
        }

        return null;
    }

    // -----------------------------------------------------------------------

    public static String getAttribute(Element element,
                                      String namespace, String localName) {
        String value = element.getAttributeNS(namespace, localName);
        if (value == null || (value = value.trim()).length() == 0) {
            value = null;
        }
        return value;
    }

    public static boolean getBooleanAttribute(Element element,
                                              String namespace, 
                                              String localName,
                                              boolean fallback) {
        boolean result = fallback;

        String value = getAttribute(element, namespace, localName);
        if (value != null) {
            if ("true".equals(value)) {
                result = true;
            } else if ("false".equals(value)) {
                result = false;
            }
        }

        return result;
    }

    public static String getXHTMLLang(Element element) {
        String value = getAttribute(element, XML_NS_URI, "lang");
        if (value == null) {
            value = getAttribute(element, null, "lang");
        }
        return value;
    }
}
