/*
 * Copyright (c) 2017 XMLmind Software. All rights reserved.
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind Web Help Compiler project.
 * For conditions of distribution and use, see the accompanying LEGAL.txt file.
 */
package com.xmlmind.whc;

import java.util.Locale;
import static javax.xml.XMLConstants.XML_NS_URI;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import com.xmlmind.util.ArrayUtil;
import com.xmlmind.util.XMLText;

/*package*/ final class IndexEntry {
    public final IndexTerm term;
    public final String id; // May be null.

    public IndexRef[] seeList; // May be null.
    public IndexAnchor[] anchorList; // May be null.
    public IndexRef[] seeAlsoList; // May be null.
    public IndexEntry[] entryList; // May be null.

    // -----------------------------------------------------------------------

    public static IndexEntry parse(Element entryElement) 
        throws IllegalArgumentException {
        IndexEntry entry = null;

        String id = DOMUtil.getAttribute(entryElement, XML_NS_URI, "id");

        Node node = entryElement.getFirstChild();
        while (node != null) {
            if (node.getNodeType() == Node.ELEMENT_NODE) {
                Element child = (Element) node;
                String localName = node.getLocalName();

                if ("term".equals(localName)) {
                    if (entry != null) {
                        throw new IllegalArgumentException("");
                    }

                    String termText = 
                        XMLText.collapseWhiteSpace(node.getTextContent());
                    if (termText.length() == 0) {
                        throw new IllegalArgumentException(
                            Msg.msg("voidElement", "term"));
                    }

                    String termHTML = SaveDocument.xhtmlToString(node);
                    
                    String sortAs = 
                        DOMUtil.getAttribute(child, null, "sortAs");
                    if (sortAs != null) {
                        sortAs = XMLText.collapseWhiteSpace(sortAs);
                    }

                    entry = 
                      new IndexEntry(new IndexTerm(termHTML, termText, sortAs),
                                     id);
                } else if ("see".equals(localName) ||
                           "seeAlso".equals(localName)) {
                    checkEntry(entry);

                    String ref = DOMUtil.getAttribute(child, null, "ref");

                    String termText = 
                        XMLText.collapseWhiteSpace(node.getTextContent());
                    String termHTML = null;
                    if (termText.length() > 0) {
                        termHTML = SaveDocument.xhtmlToString(node);
                    } else {
                        termText = null;
                    }

                    if (ref == null && termText == null) {
                        throw new IllegalArgumentException(
                            Msg.msg("voidElement", localName));
                    }

                    IndexRef see = new IndexRef(ref, termHTML, termText);
                    if ("see".equals(localName)) {
                        entry.addSee(see);
                    } else {
                        entry.addSeeAlso(see);
                    }
                } else if ("anchor".equals(localName)) {
                    checkEntry(entry);

                    String href = getRequiredAttribute(child, "href");
                    String href2 = DOMUtil.getAttribute(child, null, "href2");

                    entry.addAnchor(new IndexAnchor(href, href2));
                } else if ("entry".equals(localName)) {
                    checkEntry(entry);

                    IndexEntry subEntry = parse(child);
                    entry.addEntry(subEntry);
                }
            }

            node = node.getNextSibling();
        }

        checkEntry(entry);
        // Allow an entry with just a term. 
        // May be it will be merged with something useful.

        return entry;
    }

    private static void checkEntry(IndexEntry entry) 
        throws IllegalArgumentException {
        if (entry == null) {
            throw new IllegalArgumentException("");
        }
    }

    private static String getRequiredAttribute(Element element, 
                                               String localName) 
        throws IllegalArgumentException {
        String value = DOMUtil.getAttribute(element, null, localName);
        if (value == null) {
            throw new IllegalArgumentException(
                Msg.msg("missingAttribute", localName));
        }
        return value;
    }

    private IndexEntry(IndexTerm term, String id) {
        this.term = term;
        this.id = id;
    }

    private void addSee(IndexRef ref) {
        if (seeList == null) {
            seeList =  new IndexRef[] { ref };
        } else {
            if (ArrayUtil.indexOf(seeList, ref) < 0) {
                seeList = ArrayUtil.append(seeList, ref);
            }
        }
    }

    private void addAnchor(IndexAnchor anchor) {
        if (anchorList == null) {
            anchorList =  new IndexAnchor[] { anchor };
        } else {
            if (ArrayUtil.indexOf(anchorList, anchor) < 0) {
                anchorList = ArrayUtil.append(anchorList, anchor);
            }
        }
    }

    private void addSeeAlso(IndexRef ref) {
        if (seeAlsoList == null) {
            seeAlsoList =  new IndexRef[] { ref };
        } else {
            if (ArrayUtil.indexOf(seeAlsoList, ref) < 0) {
                seeAlsoList = ArrayUtil.append(seeAlsoList, ref);
            }
        }
    }

    private void addEntry(IndexEntry subEntry) {
        if (entryList == null) {
            entryList =  new IndexEntry[] { subEntry };
        } else {
            entryList = ArrayUtil.append(entryList, subEntry);
        }
    }

    public void merge(IndexEntry other, boolean caseSensitive, Locale locale) {
        IndexRef[] refs = other.seeList;
        if (refs != null) {
            for (IndexRef ref : refs) {
                addSee(ref);
            }
        }

        IndexAnchor[] anchors = other.anchorList;
        if (anchors != null) {
            for (IndexAnchor anchor : anchors) {
                addAnchor(anchor);
            }
        }

        refs = other.seeAlsoList;
        if (refs != null) {
            for (IndexRef ref : refs) {
                addSeeAlso(ref);
            }
        }

        IndexEntry[] subEntries2 = other.entryList;
        if (subEntries2 != null) {
            IndexEntry[] subEntries = entryList;
            if (subEntries == null) {
                entryList = subEntries2;
            } else {
                for (IndexEntry subEntry2 : subEntries2) {
                    String termText2 = subEntry2.term.text;
                    if (!caseSensitive) {
                        termText2 = termText2.toLowerCase(locale);
                    }

                    IndexEntry dstEntry = null;

                    for (IndexEntry subEntry : subEntries) {
                        String termText = subEntry.term.text;
                        if (!caseSensitive) {
                            termText = termText.toLowerCase(locale);
                        }

                        if (termText2.equals(termText)) {
                            dstEntry = subEntry;
                            break;
                        }
                    }

                    if (dstEntry == null) {
                        addEntry(subEntry2);
                    } else {
                        dstEntry.merge(subEntry2, caseSensitive, locale);
                    }
                }
            }
        }
    }
}
