/*
 * Copyright (c) 2017 XMLmind Software. All rights reserved.
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind Web Help Compiler project.
 * For conditions of distribution and use, see the accompanying LEGAL.txt file.
 */
package com.xmlmind.whc;

import java.io.IOException;
import java.io.Writer;
import java.util.Stack;
import static javax.xml.XMLConstants.XML_NS_URI;
import com.xmlmind.util.StringList;

/*package*/ final class NamespaceHelper {
    public static final String XHTML_NS_URI = "http://www.w3.org/1999/xhtml";
    public static final String SVG_NS_URI = "http://www.w3.org/2000/svg";
    public static final String XLINK_NS_URI = "http://www.w3.org/1999/xlink";
    public static final String XLINK_PREFIX = "xlink";
    public static final String MATHML_NS_URI = 
        "http://www.w3.org/1998/Math/MathML";
    public static final String OPS_NS_URI = "http://www.idpf.org/2007/ops";
    public static final String OPS_PREFIX = "epub";

    private int prefixCounter;
    private Stack<String[]> contextStack;

    // -----------------------------------------------------------------------

    public NamespaceHelper() {
        prefixCounter = 0;

        contextStack = new Stack<String[]>();
        contextStack.push(new String[] { "", XHTML_NS_URI });
    }

    public void pushContext() {
        contextStack.push(StringList.EMPTY_LIST);
    }

    public void popContext() {
        contextStack.pop();
    }

    public String getQName(String nsURI, String localName, String prefix, 
                           boolean isAttribute) {
        if (nsURI == null) {
            nsURI = "";
        }

        String usePrefix = null;
        if ("".equals(nsURI) && isAttribute) {
            usePrefix = "";
        } else if (XML_NS_URI.equals(nsURI)) {
            usePrefix = "xml";
        }

        if (usePrefix == null) {
            usePrefix = findPrefix(nsURI, isAttribute);
            if (usePrefix == null) {
                String candidatePrefix = null;
                boolean isRoot = false;

                if (OPS_NS_URI.equals(nsURI)) {
                    candidatePrefix = OPS_PREFIX;
                } else {
                    if (isAttribute) {
                        if (XLINK_NS_URI.equals(nsURI)) {
                            candidatePrefix = XLINK_PREFIX;
                        }
                    } else {
                        if ((XHTML_NS_URI.equals(nsURI) && 
                             "html".equals(localName)) ||
                            (SVG_NS_URI.equals(nsURI) && 
                             "svg".equals(localName))||
                            (MATHML_NS_URI.equals(nsURI) && 
                             "math".equals(localName))) {
                            candidatePrefix = "";
                            isRoot = true;
                        }
                    }
                }
                if (candidatePrefix == null) {
                    if ("".equals(prefix) && !isAttribute) {
                        prefix = null;
                    }
                    candidatePrefix = prefix;
                }

                if (!isRoot &&
                    candidatePrefix != null &&
                    findURI(candidatePrefix) != null) {
                    candidatePrefix = null;
                }

                if (candidatePrefix == null) {
                    usePrefix = "ns";
                    ++prefixCounter;
                    if (prefixCounter >= 2) {
                        usePrefix += Integer.toString(prefixCounter);
                    }
                } else {
                    usePrefix = candidatePrefix;
                }

                String[] bindings = contextStack.peek();
                int bindingCount = bindings.length;
                String[] bindings2 = new String[bindingCount+2];
                System.arraycopy(bindings, 0, bindings2, 0, bindingCount);
                bindings2[bindingCount] = usePrefix;
                bindings2[bindingCount+1] = nsURI;
                
                contextStack.set(contextStack.size()-1, bindings2);
            }
        }

        return ((usePrefix.length() == 0)? 
                localName : usePrefix + ":" + localName);
    }

    private String findPrefix(String nsURI, boolean isAttribute) {
        for (int j = contextStack.size()-1; j >= 0; --j) {
            String[] bindings = contextStack.get(j);

            final int count = bindings.length;
            for (int i = 0; i < count; i += 2) {
                if (bindings[i+1].equals(nsURI)) {
                    String prefix = bindings[i];
                    if (!isAttribute || prefix.length() > 0) {
                        return prefix;
                    }
                }
            }
        }

        return null;
    }

    private String findURI(String prefix) {
        for (int j = contextStack.size()-1; j >= 0; --j) {
            String[] bindings = contextStack.get(j);

            final int count = bindings.length;
            for (int i = 0; i < count; i += 2) {
                if (bindings[i].equals(prefix)) {
                    return bindings[i+1];
                }
            }
        }

        return null;
    }

    public void declarePrefixes(Writer out) 
        throws IOException {
        String[] bindings = contextStack.peek();
        final int count = bindings.length;
        for (int i = 0; i < count; i += 2) {
            String prefix = bindings[i];
            String nsURI = bindings[i+1];

            out.write(" xmlns");
            if (prefix.length() > 0) {
                out.write(':');
                out.write(prefix);
            }
            out.write("=\"");
            saveText(nsURI, out);
            out.write('"');
        }
    }

    public static void saveText(String text, Writer out)  
        throws IOException {
        final int count = text.length();
        for (int i = 0; i < count; ++i) {
            char c = text.charAt(i);
            switch (c) {
            case '\"':
                out.write("&quot;");
                break;
            case '<':
                out.write("&lt;");
                break;
            case '>':
                out.write("&gt;");
                break;
            case '&':
                out.write("&amp;");
                break;
            default:
                out.write(c);
            }
        }
    }
}
