/*
 * Copyright (c) 2017 XMLmind Software. All rights reserved.
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind Web Help Compiler project.
 * For conditions of distribution and use, see the accompanying LEGAL.txt file.
 */
package com.xmlmind.whc;

import java.util.Stack;
import org.xml.sax.SAXException;
import org.xml.sax.Locator;
import org.xml.sax.Attributes;
import org.xml.sax.ext.DefaultHandler2;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Document;

/*package*/ final class SAXToDOM extends DefaultHandler2 {
    public final Document doc;

    private Stack<Node> nodeStack;
    private StringBuilder buffer;
    private Locator locator;

    // -----------------------------------------------------------------------

    public SAXToDOM(Document doc) {
        this.doc = doc;

        nodeStack = new Stack<Node>();
        nodeStack.push(doc);
    }

    @Override
    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    @Override
    public void startDTD(String name, String publicId, String systemId) {
        DocumentType docType = 
          doc.getImplementation().createDocumentType(name, publicId, systemId);
        doc.appendChild(docType);
    }

    @Override
    public void startElement(String uri, String localName, String qName,
                             Attributes atts)
        throws SAXException {
        flushBuffer();

        Node parent = nodeStack.peek();

        Element element = createElement(parent, uri, qName, atts);

        parent.appendChild(element);

        nodeStack.push(element);
    }

    private Element createElement(Node parent, String uri, String qName,
                                  Attributes atts) {
        Element element = 
            doc.createElementNS(((uri == null || uri.length() == 0)? 
                                 null : uri), 
                                qName);

        int attCount = atts.getLength();
        for (int i = 0; i < attCount; ++i) {
            String attName = atts.getQName(i);
            if (attName.startsWith("xmlns")) {
                continue;
            }

            String attURI = atts.getURI(i);

            element.setAttributeNS(((attURI == null || attURI.length() == 0)? 
                                    null : attURI), 
                                   attName, atts.getValue(i));
        }

        NodeLocation location = null;
        if (locator != null) {
            String systemId = locator.getSystemId();
            if (systemId != null) {
                location = new NodeLocation(systemId,
                                            locator.getLineNumber(), 
                                            locator.getColumnNumber());
            }
        }
        if (location != null) {
            element.setUserData(NodeLocation.USER_DATA_KEY, location,
                                DOMUtil.COPY_USER_DATA);
        }

        return element;
    }

    @Override
    public void endElement(String uri, String localName, String qName) {
        flushBuffer();

        nodeStack.pop();
    }

    private void flushBuffer() {
        if (buffer != null) {
            Node parent = nodeStack.peek();
            parent.appendChild(doc.createTextNode(buffer.toString()));
            buffer = null;
        }
    }

    @Override
    public void characters(char[] ch, int start, int length) {
        if (buffer == null) {
            buffer = new StringBuilder();
        }
        buffer.append(ch, start, length);
    }

    @Override
    public void ignorableWhitespace(char[] ch, int start, int length) {
        if (buffer == null) {
            buffer = new StringBuilder();
        }
        buffer.append(ch, start, length);
    }
}
