/*
 * Copyright (c) 2017-2019 XMLmind Software. All rights reserved.
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind Web Help Compiler project.
 * For conditions of distribution and use, see the accompanying LEGAL.txt file.
 */
package com.xmlmind.whc;

import java.io.IOException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.Writer;
import java.io.OutputStreamWriter;
import java.io.BufferedWriter;
import java.io.StringWriter;
import java.util.HashSet;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.Text;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Element;
import org.w3c.dom.Attr;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Document;
import com.xmlmind.util.ThrowableUtil;
import com.xmlmind.util.XMLText;
import com.xmlmind.util.XMLUtil;
import static com.xmlmind.whc.NamespaceHelper.XHTML_NS_URI;

/**
 * <b>Not part of the public documented API:</b> helper class used to
 * save XHTML documents.
 */
public final class SaveDocument {
    private SaveDocument() {}

    // -----------------------------------------------------------------------

    public static void saveXHTML(Document doc, File outFile) 
        throws IOException {
        saveXHTML(doc, outFile, "UTF-8", false);
    }

    public static void saveXHTML(Document doc, File outFile, String charset) 
        throws IOException {
        saveXHTML(doc, outFile, charset, false);
    }

    public static void saveXHTML(Document doc, File outFile, String charset, 
                                 boolean savePI) 
        throws IOException {
        // A /html/head/meta is assumed to have been added to specify charset.

        Writer out = new BufferedWriter(
            new OutputStreamWriter(new FileOutputStream(outFile), charset));
        try {
            saveXHTML(doc, savePI, out);
            out.flush();
        } finally {
            out.close();
        }
    }

    private static void saveXHTML(Document doc, boolean savePI, Writer out) 
        throws IOException {
        DocumentType docType = doc.getDoctype();
        if (docType != null) {
            saveDocType(docType, out);
        }

        NamespaceHelper nsHelper = new NamespaceHelper();
        nsHelper.pushContext();

        out.write("<html xmlns=\"");
        out.write(XHTML_NS_URI);
        out.write("\"");

        Element root = doc.getDocumentElement();
        saveAttributes(root, nsHelper, out);
        out.write('>');

        saveContents(root, savePI, nsHelper, out);

        out.write("</html>");

        nsHelper.popContext();
    }

    private static void saveDocType(DocumentType docType, Writer out) 
        throws IOException {
        out.write("<!DOCTYPE ");
        out.write(docType.getName());

        String systemId = docType.getSystemId();
        if (systemId != null) {
            String publicId = docType.getPublicId();
            if (publicId == null) {
                out.write(" SYSTEM ");
            } else {
                // publicId cannot contain double-quotes, nor markup.
                out.write(" PUBLIC \"");
                out.write(publicId);
                out.write("\"\n");
            }

            char quote = (systemId.indexOf('\"') < 0)? '\"' : '\'';
            out.write(quote);
            // systemId cannot contain markup.
            out.write(systemId);
            out.write(quote);
        }

        String internalSubset = docType.getInternalSubset();
        if (internalSubset != null) {
            internalSubset = internalSubset.trim();
            if (internalSubset.length() == 0) {
                internalSubset = null;
            }
        }
        if (internalSubset != null) {
            out.write(" [\n");
            out.write(internalSubset);
            out.write("\n]");
        }

        out.write(">\n");
    }

    private static void saveContents(Node tree, boolean savePI, 
                                     NamespaceHelper nsHelper, Writer out) 
        throws IOException {
        Node node = tree.getFirstChild();
        while (node != null) {
            switch (node.getNodeType()) {
            case Node.TEXT_NODE:
                NamespaceHelper.saveText(((Text) node).getData(), out);
                break;

            case Node.PROCESSING_INSTRUCTION_NODE:
                if (savePI) {
                    ProcessingInstruction pi = (ProcessingInstruction) node;
                    String data = pi.getData();
                    
                    out.write("<?");
                    out.write(pi.getTarget());
                    if (data.length() > 0) {
                        out.write(' ');
                        out.write(data);
                    }
                    out.write("?>");
                }
                break;

            case Node.ELEMENT_NODE:
                saveElement((Element) node, savePI, nsHelper, out);
                break;
            }

            node = node.getNextSibling();
        }
    }

    private static void saveElement(Element element, boolean savePI, 
                                    NamespaceHelper nsHelper, Writer out)  
        throws IOException {
        nsHelper.pushContext();

        out.write('<');
        String ns = element.getNamespaceURI();
        String localName = element.getLocalName();
        String tag = nsHelper.getQName(ns, localName, element.getPrefix(), 
                                       /*attr*/ false);
        out.write(tag);

        saveAttributes(element, nsHelper, out);

        if (XMLUtil.HTML_NO_END_TAG.contains(tag)) {
            out.write(" />");
        } else {
            out.write('>');

            if (("script".equals(localName) || "style".equals(localName)) &&
                XHTML_NS_URI.equals(ns)) {
                saveCDATA(element.getTextContent(), out);
            } else {
                saveContents(element, savePI, nsHelper, out);
            }

            out.write("</");
            out.write(tag);
            out.write('>');
        }

        nsHelper.popContext();
    }

    private static void saveAttributes(Element element,
                                       NamespaceHelper nsHelper, Writer out)
        throws IOException {
        NamedNodeMap attrs = element.getAttributes();

        final int attrCount = attrs.getLength();
        for (int i = 0; i < attrCount; ++i) {
            Attr attr = (Attr) attrs.item(i);

            out.write(' ');
            String attrQName = nsHelper.getQName(attr.getNamespaceURI(),
                                                 attr.getLocalName(),
                                                 attr.getPrefix(), 
                                                 /*attr*/ true);
            out.write(attrQName);
            out.write("=\"");
            NamespaceHelper.saveText(attr.getValue(), out);
            out.write('"');
        }

        nsHelper.declarePrefixes(out);
    }

    private static void saveCDATA(String text, Writer out) 
        throws IOException {
        if (text != null && text.length() > 0) {
            if (XMLText.isXMLSpace(text)) {
                // Typically an open line.
                out.write(text);
            } else {
                out.write("/*<![CDATA[*/");
                out.write(text);
                out.write("/*]]>*/");
            }
        }
    }

    // -----------------------------------------------------------------------

    public static String xhtmlToString(Node xhtmlContainer) {
        return xhtmlToString(xhtmlContainer, false);
    }

    public static String xhtmlToString(Node xhtmlContainer, boolean savePI) {
        NamespaceHelper nsHelper = new NamespaceHelper();
        nsHelper.pushContext();

        StringWriter out = new StringWriter();
        try {
            saveContents(xhtmlContainer, savePI, nsHelper, out);
        } catch (IOException cannotHappen) {}

        nsHelper.popContext();

        return out.toString();
    }
}
