/*
 * Copyright (c) 2017-2019 XMLmind Software. All rights reserved.
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind Web Help Compiler project.
 * For conditions of distribution and use, see the accompanying LEGAL.txt file.
 */
package com.xmlmind.whc;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;

/*package*/ final class SearchInfo {
    public static final class WordRef implements Comparable<WordRef> {
        public final int pageIndex;
        public int refCount;

        public WordRef(int pageIndex) {
            this.pageIndex = pageIndex;
            refCount = 1;
        }

        public int compareTo(WordRef other) {
            // WordRef are sorted first by the number of references per page
            // (more references before less references) and then by the
            // increasing index of the page.
            int delta = (other.refCount - refCount);
            if (delta == 0) {
                return (pageIndex - other.pageIndex);
            } else {
                return delta;
            }
        }
    }

    public ArrayList<String> baseNameList;
    public ArrayList<String> titleList;
    public HashMap<String,WordRef[]> wordMap;

    // -----------------------------------------------------------------------

    public SearchInfo() {
        baseNameList = new ArrayList<String>();
        titleList = new ArrayList<String>();
        wordMap = new HashMap<String,WordRef[]>();
    }

    public String toJS(String stemmerJS) {
        StringBuilder buffer = new StringBuilder();

        if (stemmerJS != null) {
            buffer.append(stemmerJS);
            buffer.append("\n\n");
        }

        // ---

        buffer.append("wh.search_stemmer = ");
        if (stemmerJS != null) {
            buffer.append("Snowball()");
        } else {
            buffer.append("null");
        }
        buffer.append(";\n");

        // ---

        buffer.append("wh.search_baseNameList = [\n");

        int last = baseNameList.size()-1;
        for (int i = 0; i <= last; ++i) {
            String baseName = baseNameList.get(i);

            buffer.append("/*");
            buffer.append(Integer.toString(i));
            buffer.append("*/ ");

            Util.appendJSString(baseName, buffer);
            if (i != last) {
                buffer.append(",\n");
            } else {
                buffer.append('\n');
            }
        }

        buffer.append("];\n");

        // ---

        buffer.append("wh.search_titleList = [\n");

        last = titleList.size()-1;
        for (int i = 0; i <= last; ++i) {
            String title = titleList.get(i);

            buffer.append("/*");
            buffer.append(Integer.toString(i));
            buffer.append("*/ ");

            Util.appendJSString(title, buffer);
            if (i != last) {
                buffer.append(",\n");
            } else {
                buffer.append('\n');
            }
        }

        buffer.append("];\n");

        // ---

        int i = 0;
        last = wordMap.size()-1;

        buffer.append("// ");
        buffer.append(Integer.toString(last+1));
        buffer.append(" words.\n");

        buffer.append("wh.search_wordMap= {\n");

        Iterator<Map.Entry<String,WordRef[]>> iter = 
            wordMap.entrySet().iterator();
        while (iter.hasNext()) {
            Map.Entry<String,WordRef[]> e = iter.next();

            String word = e.getKey();
            WordRef[] refs = e.getValue();

            Util.appendJSString(word, buffer);
            buffer.append(": [");
                
            Arrays.sort(refs);
            appendWordRefs(refs, buffer);

            if (i != last) {
                buffer.append("],\n");
            } else {
                buffer.append("]\n");
            }
            ++i;
        }

        buffer.append("};\n");

        return buffer.toString();
    }

    private static void appendWordRefs(WordRef[] refs, StringBuilder buffer) {
        int from = -1;
        int to = -1;
        int prevRefCount = -1;
        boolean firstItem = true;
        
        final int count = refs.length;
        for (int i = 0; i < count; ++i) {
            WordRef ref = refs[i];

            if (ref.refCount == prevRefCount) {
                to = i;
            } else {
                if (from >= 0) {
                    if (firstItem) {
                        firstItem = false;
                    } else {
                        buffer.append(',');
                    }

                    appendItem(refs, from, to, buffer);
                }

                from = to = i;
                prevRefCount = ref.refCount;
            }
        }

        if (from >= 0) {
            if (firstItem) {
                firstItem = false;
            } else {
                buffer.append(',');
            }

            appendItem(refs, from, to, buffer);
        }
    }

    private static void appendItem(WordRef[] refs, int from, int to,
                                   StringBuilder buffer) {
        if (from == to) {
            buffer.append(Integer.toString(refs[from].pageIndex));
        } else {
            buffer.append('[');
            for (int i = from; i <= to; ++i) {
                if (i > from) {
                    buffer.append(',');
                }
                buffer.append(Integer.toString(refs[i].pageIndex));
            }
            buffer.append(']');
        }
    }
}
