/*
 * Copyright (c) 2017-2019 XMLmind Software. All rights reserved.
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind Web Help Compiler project.
 * For conditions of distribution and use, see the accompanying LEGAL.txt file.
 */
package com.xmlmind.whc;

// Needed by main.
import java.io.IOException;
import java.io.File;
import com.xmlmind.util.ThrowableUtil;
import com.xmlmind.util.FileUtil;

/*package*/ final class Util {
    private Util() {}

    public static final String EMACS_UTF8_DECL = 
        "// -*- coding: utf-8 -*-\n\n";

    // -----------------------------------------------------------------------

    public static void appendJSString(String string, StringBuilder buffer) {
        buffer.append('"');

	final int length = string.length();
	for (int i = 0; i < length; ++i) {
	    char c = string.charAt(i);
	    switch (c) {
	    case '\b':
		buffer.append("\\b");
		break;
	    case '\f':
		buffer.append("\\f");
		break;
	    case '\n':
		buffer.append("\\n");
		break;
	    case '\r':
		buffer.append("\\r");
		break;
	    case '\t':
		buffer.append("\\t");
		break;
	    case '\'':
		buffer.append("\\\'");
		break;
	    case '"':
		buffer.append("\\\"");
		break;
	    case '\\':
		buffer.append("\\\\");
		break;
	    default:
		buffer.append(c);
	    }
	}

        buffer.append('"');
    }

    // -----------------------------------------------------------------------

    public static String compactJS(String source) {
        return compact(source, /*cssMode*/ false);
    }

    public static String compactCSS(String source) {
        return compact(source, /*cssMode*/ true);
    }

    private enum State {
        MULTI_LINE_COMMENT,
        LINE_COMMENT,  /*Not used in cssMode*/
        STRING_QUOTE,
        STRING_APOS,
        SPACE_CHARS,
        NORMAL_CHARS
    };

    private static String compact(String source, boolean cssMode) {
        final char[] in = source.toCharArray();
        final int inLength = in.length;

        char[] out = new char[inLength];
        int outLength = 0;

        State state = State.NORMAL_CHARS;

        for (int i = 0; i < inLength; ++i) {
            char c = in[i];

            switch (c) {
            case '/':
                switch (state) {
                case MULTI_LINE_COMMENT:
                    if (i > 0 && in[i-1] == '*') {
                        state = State.NORMAL_CHARS;
                    }
                    // Otherwise, still inside a multi-line comment.
                    break;
                case LINE_COMMENT:
                    // Skip comment.
                    break;
                case STRING_QUOTE:
                case STRING_APOS:
                    out[outLength++] = c;
                    break;
                case SPACE_CHARS:
                case NORMAL_CHARS:
                    if (i+1 < inLength) {
                        char nextC = in[i+1];
                        if (nextC == '*') {
                            ++i;
                            state = State.MULTI_LINE_COMMENT;
                        } else if (nextC == '/' && !cssMode) {
                            ++i;
                            state = State.LINE_COMMENT;
                        } else {
                            if (!cssMode) {
                                // JavaScript regexp?
                                int j = lastREOffset(in, i, inLength);
                                if (j > i) {
                                    // Possibly a regexp.
                                    assert(in[j] == '/');
                                    int copied = j - i + 1;
                                    System.arraycopy(in, i, out, outLength, 
                                                     copied);
                                    outLength += copied;
                                    i = j;
                                } else {
                                    // Definitely not a regexp.
                                    out[outLength++] = c;
                                }
                            } else {
                                out[outLength++] = c;
                            }
                            state = State.NORMAL_CHARS;
                        }
                    } else {
                        out[outLength++] = c;
                        state = State.NORMAL_CHARS;
                    }
                    break;
                }
                break;

            case '"':
                switch (state) {
                case MULTI_LINE_COMMENT:
                case LINE_COMMENT:
                    // Skip comment.
                    break;
                case STRING_QUOTE:
                    // End of string.
                    out[outLength++] = c;
                    state = State.NORMAL_CHARS;
                    break;
                case STRING_APOS:
                    out[outLength++] = c;
                    break;
                case SPACE_CHARS:
                case NORMAL_CHARS:
                    out[outLength++] = c;
                    state = State.STRING_QUOTE;
                    break;
                }
                break;

            case '\'':
                switch (state) {
                case MULTI_LINE_COMMENT:
                case LINE_COMMENT:
                    // Skip comment.
                    break;
                case STRING_QUOTE:
                    out[outLength++] = c;
                    break;
                case STRING_APOS:
                    // End of string.
                    out[outLength++] = c;
                    state = State.NORMAL_CHARS;
                    break;
                case SPACE_CHARS:
                case NORMAL_CHARS:
                    out[outLength++] = c;
                    state = State.STRING_APOS;
                    break;
                }
                break;

            case '\n':
                switch (state) {
                case MULTI_LINE_COMMENT:
                    // Skip comment.
                    break;
                case STRING_QUOTE:
                case STRING_APOS:
                    // Probably a syntax error in the source.
                    out[outLength++] = c;
                    state = State.SPACE_CHARS;
                    break;
                case LINE_COMMENT:
                case SPACE_CHARS:
                case NORMAL_CHARS:
                    if (outLength > 0 && out[outLength-1] != '\n') {
                        out[outLength++] = c;
                    }
                    // Otherwise, skip open line.
                    state = State.SPACE_CHARS;
                    break;
                }
                break;

            case ' ': case '\t': case '\r': case '\f': case '\u000B': /*'\v'*/
                switch (state) {
                case MULTI_LINE_COMMENT:
                case LINE_COMMENT:
                    // Skip comment.
                    break;
                case STRING_QUOTE:
                case STRING_APOS:
                    out[outLength++] = c;
                    break;
                case SPACE_CHARS:
                    // Compress whitespace.
                    break;
                case NORMAL_CHARS:
                    out[outLength++] = ' '; // Normalized to a space char.
                    state = State.SPACE_CHARS;
                    break;
                }
                break;

            case '\\':
                switch (state) {
                case MULTI_LINE_COMMENT:
                case LINE_COMMENT:
                    // Skip comment.
                    break;
                case STRING_QUOTE:
                case STRING_APOS:
                    if (i+1 < inLength) {
                        char nextC = in[i+1];
                        if (nextC == '"' || nextC == '\'') {
                            // Escaped quote or apostrophe.
                            // (Always supported by both JavaScript and CSS.)
                            out[outLength++] = c;
                            out[outLength++] = nextC;
                            ++i;
                        } else if (nextC == '\n' && !cssMode) {
                            // Escaped newline.
                            // (A CSS string cannot directly contain a newline.)
                            out[outLength++] = c;
                            out[outLength++] = nextC;
                            ++i;
                        } else {
                            out[outLength++] = c;
                        }
                    } else {
                        out[outLength++] = c;
                    }
                    break;
                case SPACE_CHARS:
                    // End of whitespace.
                    out[outLength++] = c;
                    state = State.NORMAL_CHARS;
                    break;
                case NORMAL_CHARS:
                    out[outLength++] = c;
                    break;
                }
                break;

            default:
                switch (state) {
                case MULTI_LINE_COMMENT:
                case LINE_COMMENT:
                    // Skip comment.
                    break;
                case SPACE_CHARS:
                    // End of whitespace.
                    out[outLength++] = c;
                    state = State.NORMAL_CHARS;
                    break;
                case STRING_QUOTE:
                case STRING_APOS:
                case NORMAL_CHARS:
                    out[outLength++] = c;
                    break;
                }
            }
        }

        /*
        System.err.println("Util.compact>>> " + inLength + " chars --> " + 
                           outLength + " chars");
        */

        return new String(out, 0, outLength);
    }
    
    private static int lastREOffset(char[] in, int i, int inLength) {
        assert(in[i] == '/');

        for (int j = i+1; j < inLength; ++j) {
            switch (in[j]) {
            case '\\':
                if (j+1 < inLength && in[j+1] == '/') {
                    // Skip escaped slash.
                    ++j;
                }
                break;

            case '/':
                if (j+1 < inLength) {
                    char nextC = in[j+1];
                    if (nextC == '*' || nextC == '/') {
                        // Found comment. Hence not a regexp.
                        return -1;
                    } else {
                        return j;
                    }
                } else {
                    return j;
                }
                /*NOTREACHED*/

            case '\n':
                // Found EOL. Hence not a regexp.
                return -1;
            }
        }

        return -1;
    }

    public static void main(String[] args) {
        File inFile = null;;
        if (args.length != 2 ||
            !(inFile = new File(args[0])).isFile()) {
            System.err.println(
                "Usage: java -cp whc.jar com.xmlmind.whc.Util" +
                " in_css_or_js_file out_file");
            System.exit(1);
        }

        File outFile = new File(args[1]);

        try {
            String source = FileUtil.loadString(inFile, "UTF-8");

            String minified = compact(source, args[0].endsWith(".css"));

            FileUtil.saveString(minified, outFile, "UTF-8");
        } catch (IOException e) {
            System.err.println("Cannot compact '" + inFile + "': " + 
                               ThrowableUtil.reason(e));
            System.exit(2);
        }
    }
}
