/*
 * Copyright (c) 2017 XMLmind Software. All rights reserved.
 *
 * Author: Hussein Shafie
 *
 * This file is part of the XMLmind Web Help Compiler project.
 * For conditions of distribution and use, see the accompanying LEGAL.txt file.
 */
package com.xmlmind.whc;

import java.util.Map;
import java.util.Stack;
import java.util.HashMap;
import java.util.Locale;
import java.util.Stack;
import java.text.BreakIterator;
import org.w3c.dom.Node;
import org.w3c.dom.Text;
import org.w3c.dom.Element;
import com.xmlmind.util.LocaleUtil;
import com.xmlmind.util.XMLText;
import static com.xmlmind.whc.NamespaceHelper.XHTML_NS_URI;

/*package*/ final class WordCollector {
    private HashMap<Locale,BreakIterator> localeToWordBreakIter;

    private Stack<String> langStack;

    // -----------------------------------------------------------------------

    public WordCollector() {
        localeToWordBreakIter = new HashMap<Locale,BreakIterator>();
    }

    public void collect(Element tree, String initialLang, int pageIndex,
                        Stemmer stemmer, boolean indexNumbers,
                        Map<String,SearchInfo.WordRef[]> wordMap) {
        langStack = new Stack<String>();
        langStack.push(initialLang);

        doCollect(tree, pageIndex, stemmer, indexNumbers, wordMap);

        langStack.pop();
    }

    private void doCollect(Element tree, int pageIndex, 
                           Stemmer stemmer, boolean indexNumbers,
                           Map<String,SearchInfo.WordRef[]> wordMap) {
        if (!XHTML_NS_URI.equals(tree.getNamespaceURI())) {
            // Do not index text found inside embedded SVG, embedded MathML or
            // "XML islands".
            return;
        }

        // ---

        String lang = DOMUtil.getXHTMLLang(tree);
        if (lang == null) {
            lang = langStack.peek();
        }
        langStack.push(lang);

        Locale locale = null;
        BreakIterator wordBreakIter = null;

        Node node = tree.getFirstChild();
        while (node != null) {
            switch (node.getNodeType()) {
            case Node.TEXT_NODE:
                {
                    String text = ((Text) node).getData();
                    if (!XMLText.isXMLSpace(text)) {
                        if (locale == null) {
                            locale = LocaleUtil.getLocale(lang);
                            wordBreakIter = getWordBreakIter(locale);
                        }
                        collectWords(text, locale, wordBreakIter, 
                                     pageIndex, stemmer, indexNumbers, wordMap);
                    }
                }
                break;

            case Node.ELEMENT_NODE:
                doCollect((Element) node, pageIndex, stemmer, indexNumbers,
                          wordMap);
                break;
            }

            node = node.getNextSibling();
        }

        langStack.pop();
    }

    private BreakIterator getWordBreakIter(Locale locale) {
        BreakIterator iter = localeToWordBreakIter.get(locale);
        if (iter == null) {
            iter = BreakIterator.getWordInstance(locale);
            localeToWordBreakIter.put(locale, iter);
        }
        return iter;
    }

    private static void collectWords(String text, Locale locale, 
                                     BreakIterator wordBreakIter,
                                     int pageIndex,
                                     Stemmer stemmer, boolean indexNumbers,
                                     Map<String,SearchInfo.WordRef[]> wordMap) {
        wordBreakIter.setText(text);

        String wordLang;
        if (stemmer != null &&
            ((wordLang = locale.getLanguage()) == null ||
             !wordLang.regionMatches(/*ignoreCase*/ true, 0, 
                                     stemmer.langCode, 0, 2))) {
            stemmer = null;
        }
        
        StringBuilder number = null;

        int start = wordBreakIter.first();
        for (int end = wordBreakIter.next();
             end != BreakIterator.DONE;
             start = end, end = wordBreakIter.next()) {
            String word = text.substring(start, end).toLowerCase(locale);
            int wordLength = (end - start);

            char c = text.charAt(start);
            switch (c) {
            case '-': case '+': 
                if (indexNumbers) {
                    if (number == null) {
                        number = new StringBuilder();
                    }
                    number.append(word);
                }
                break;
            default:
                if (Character.isDigit(c)) {
                    if (indexNumbers) {
                        if (number == null) {
                            number = new StringBuilder();
                        }
                        number.append(word);
                    }
                } else if (c == '.' ||
                           c == ',' ||
                           c == '%' ||
                           c == '\u2030' || // Per mille
                           Character.getType(c) == Character.CURRENCY_SYMBOL) {
                    if (number != null) {
                        number.append(word);
                    }
                    // Otherwise, ignore.
                } else if (Character.isLetter(c)) {
                    if (number != null) {
                        number.append(word); // 'px', 'cm'
                    } else {
                        if (number != null) {
                            collectNumber(number, pageIndex, wordMap);
                            number = null;
                        }

                        if (wordLength >= 2) {
                            if (stemmer != null) {
                                word = stemmer.stem(word);
                            }
                            collectWord(word, pageIndex, wordMap);
                        }
                    }
                } else {
                    // For example, a whitespace or punctuation character.
                    if (number != null) {
                        collectNumber(number, pageIndex, wordMap);
                        number = null;
                    }
                }
            }
        }

        if (number != null) {
            collectNumber(number, pageIndex, wordMap);
        }
    }

    private static void collectNumber(StringBuilder number, int pageIndex,
                                     Map<String,SearchInfo.WordRef[]> wordMap) {
        // Remove trailing '-', '+', ',' or '.' ---

        int length = number.length();
        int c;
        while (length > 0 && 
               ((c = number.charAt(length-1)) == '.' || c == ',' || 
                c == '+' || c == '-')) {
            number.deleteCharAt(length-1);
            --length;
        }

        // ---

        if (length > 0) {
            c = number.charAt(0);
            if ((c == '+' || c == '-') && 
                (length < 2 || !Character.isDigit(number.charAt(1)))) {
                // Not a number.
                return;
            }

            collectWord(number.toString(), pageIndex, wordMap);
        }
    }

    private static void collectWord(String word, int pageIndex,
                                    Map<String,SearchInfo.WordRef[]> wordMap) {
        SearchInfo.WordRef[] refs = wordMap.get(word);
        if (refs == null) {
            wordMap.put(word, 
                        new SearchInfo.WordRef[] { 
                            new SearchInfo.WordRef(pageIndex) 
                        });
        } else {
            final int count = refs.length;
            if (refs[count-1].pageIndex == pageIndex) {
                ++refs[count-1].refCount;
            } else {
                SearchInfo.WordRef[] refs2 = new SearchInfo.WordRef[count+1];
                System.arraycopy(refs, 0, refs2, 0, count);
                refs2[count] = new SearchInfo.WordRef(pageIndex);

                wordMap.put(word, refs2);
            }
        }
    }
}
